#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  ---------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
from builtins import (bytes, str, range, dict)
import unittest


# --------------------------------------------------
# should not be documented with docstrings
class PATTestCase(unittest.TestCase):
    def setUp(self):
        self.T = None

    def eStr(self, code):
        import CGNS.PAT.cgnserrors as CGE
        return CGE.TAG + CGE.TAG_ERROR + "\[%.3d\].*$" % code

    def genTree(self):
        # should not self reference for test
        # then load tree
        import CGNS.PAT.test.disk
        self.T = CGNS.PAT.test.disk.T

    def test_00Module(self):
        pass

    def test_01Check(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        self.assertFalse(CGU.checkName('/name'))
        self.assertFalse(CGU.checkNodeName(3))
        self.assertFalse(CGU.checkName(3))
        self.assertFalse(CGU.checkName(['name']))
        self.assertFalse(CGU.checkName(''))
        self.assertFalse(CGU.checkName('x' * 33))
        self.assertFalse(CGU.checkName('.'))
        self.assertFalse(CGU.checkName('..'))
        self.assertFalse(CGU.checkName('         '))
        self.assertTrue(CGU.checkName('name  test'))
        self.assertTrue(CGU.checkName('name"test"'))
        self.assertTrue(CGU.checkName('name\\test'))
        self.assertTrue(CGU.checkName(' name'))
        self.assertTrue(CGU.checkName('name'))
        self.assertTrue(CGU.checkName('x' * 32))
        self.assertTrue(CGU.checkNodeName(['test', None, [], 'Zone_t']))
        self.assertTrue(CGU.checkNodeName(['  test', None, [], 'Zone_t']))
        self.assertTrue(CGU.checkNodeName(['?test#320', None, [], 'Zone_t']))
        self.assertTrue(CGU.checkNodeName(['test   ', None, [], 'Zone_t']))
        self.assertFalse(CGU.checkName('name  test', strict=True))
        self.assertFalse(CGU.checkName('name\\test', strict=True))
        self.assertFalse(CGU.checkName('name"test"', strict=True))
        self.assertFalse(CGU.checkName(' name', strict=True))
        import string
        clist = list(string.ascii_letters + string.digits + string.punctuation + ' ')
        clist.remove('/')
        for c in clist:
            self.assertTrue(CGU.checkName('_' + c))
        ex = CGE.cgnsNameError
        fn = CGU.checkName
        self.assertRaisesRegex(ex, self.eStr(2), CGU.checkNodeName, [], dienow=True)
        self.assertRaisesRegex(ex, self.eStr(22), fn, 3, dienow=True)
        self.assertRaisesRegex(ex, self.eStr(23), fn, '', dienow=True)
        self.assertRaisesRegex(ex, self.eStr(24), fn, '/', dienow=True)
        self.assertRaisesRegex(ex, self.eStr(25), fn, 'a' * 33, dienow=True)
        self.assertRaisesRegex(ex, self.eStr(29), fn, '.', dienow=True)
        self.assertRaisesRegex(ex, self.eStr(31), fn, ' ' * 5, dienow=True)
        self.assertRaisesRegex(ex, self.eStr(32), fn, ' a' * 5, dienow=True, strict=True)
        self.assertRaisesRegex(ex, self.eStr(33), fn, '"', dienow=True, strict=True)
        self.assertRaisesRegex(ex, self.eStr(34), fn, 'a  a', dienow=True, strict=True)
        self.genTree()
        self.assertTrue(CGU.checkSameNode(self.T, self.T))
        self.assertFalse(CGU.checkSameNode(self.T, [None, None, [], None]))
        self.assertRaisesRegex(CGE.cgnsNodeError, self.eStr(30), CGU.checkSameNode, self.T, [None, None, [], None],
                                dienow=True)

    def test_02Path(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        self.genTree()
        p = '/{Base#1}/{Zone-A}/ZoneBC'
        self.assertEqual(p, CGU.getPathNoRoot(p))
        self.assertEqual(p, CGU.getPathNoRoot('/CGNSTree' + p))
        self.assertEqual("/", CGU.getPathNoRoot('/CGNSTree'))
        self.assertEqual(CGU.getPathNormalize('/./{Base#1}/{Zone-A}/ZoneBC'), p)
        self.assertEqual(CGU.getPathNormalize('/./{Base#1}/{Zone-A}/ZoneBC//'), p)
        self.assertEqual(CGU.getPathToList(p), ['', '{Base#1}', '{Zone-A}', 'ZoneBC'])
        self.assertEqual(CGU.getPathToList(p, True), ['{Base#1}', '{Zone-A}', 'ZoneBC'])
        self.assertEqual(CGU.getPathAsTypes(self.T, p), [CGK.CGNSTree_ts, CGK.CGNSBase_ts, CGK.Zone_ts, CGK.ZoneBC_ts])
        zone = ['Zone', None, [['ZoneBC', None, [], CGK.ZoneBC_ts]], CGK.Zone_ts]
        p = '/Zone/ZoneBC'
        self.assertEqual(CGU.getPathAsTypes(zone, p, legacy=False), [CGK.Zone_ts, CGK.ZoneBC_ts])
        self.assertEqual(CGU.getPathAsTypes(zone, p, legacy=True), [CGK.ZoneBC_ts])

    def test_03NodeStructure(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        self.assertEqual(['N1', None, [], 'N_t'], CGU.nodeCreate('N1', None, [], 'N_t'))
        p = CGU.nodeCreate('N1', None, [], 'N_t')
        self.assertEqual(['N1', None, [], 'N_t'], CGU.nodeCreate('N1', None, [], 'N_t', p))
        self.assertEqual(['N1', None, [], 'N_t'], p[2][0])
        n = CGU.nodeCreate('N2', None, [], 'N_t', p)
        self.assertNotEqual(['N1', None, [], 'N_t'], p[2][1])
        self.assertIs(n, p[2][1])
        self.assertEqual(['N2', None, [], 'N_t'], CGU.nodeCopy(n))
        c = CGU.nodeCopy(n)
        self.assertIsNot(n, c)
        cname = 'NCopy'
        c = CGU.nodeCopy(n, cname)
        self.assertEqual(c[0], cname)
        del n
        self.assertEqual(c[0], cname)

    def test_04NodeChildren(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnslib as CGL
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        A = ['A', None, [], None]
        B = ['B', None, [], None]
        C = ['C', None, [], None]
        A[2] += [B]
        CGU.setAsChild(A, C)
        self.assertTrue(CGU.hasChildName(A, 'B'))
        self.assertTrue(CGU.hasChildName(A, 'C'))
        self.assertTrue(CGU.checkDuplicatedName(A, 'D'))
        self.assertFalse(CGU.checkHasChildName(A, 'D'))
        D = CGU.nodeCreate('D', None, [], None, parent=A)
        self.assertFalse(CGU.checkDuplicatedName(A, 'D'))
        self.assertTrue(CGU.checkHasChildName(A, 'D'))
        self.assertFalse(CGU.checkNodeType(C))
        d = {'None': None, 'String': 'string value', 'Integer': 10, 'Float': 1.4}
        for n, v in d.items():
            CGL.newDataArray(A, n, v)
        for name in d:
            self.assertTrue(CGU.hasChildName(A, name))

    def test_05NodeValue(self):
        import numpy
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        self.genTree()
        v = CGU.getValueByPath(self.T, '/CGNSTree/CGNSLibraryVersion')
        self.assertEqual(v, numpy.array(3.2))
        n = ['ZoneType', numpy.array(['S', 't', 'r', 'u', 'c', 't', 'u', 'r', 'e', 'd'], dtype='S', order='C'), [],
             'ZoneType_t']
        self.assertTrue(CGU.stringValueMatches(n, 'Structured'))
        # set*AsArray
        self.assertEqual(CGU.setStringAsArray('Structured').tostring().decode('ascii'), 'Structured')
        self.assertEqual(CGU.setIntegerAsArray(1), numpy.array(1, dtype='int32'))
        self.assertTrue((CGU.setIntegerAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype='int32')).all())
        self.assertEqual(CGU.setLongAsArray(1), numpy.array(1, dtype='int64'))
        self.assertTrue((CGU.setLongAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype='int64')).all())
        self.assertEqual(CGU.setFloatAsArray(1), numpy.array(1, dtype='float32'))
        self.assertTrue((CGU.setFloatAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype='float32')).all())
        self.assertEqual(CGU.setDoubleAsArray(1), numpy.array(1, dtype='float64'))
        self.assertTrue((CGU.setDoubleAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype='float64')).all())
        self.assertTrue((CGU.setDoubleAsArray(tuple(range(10, 1010, 10))) == numpy.array(tuple(range(10, 1010, 10)),
                                                                                         dtype='float64')).all())
        n = ['ZoneType', None, [], 'ZoneType_t']
        # set*ByPath
        self.assertTrue(CGU.stringValueMatches(CGU.setStringByPath(n, '/', 'Structured'), 'Structured'))
        self.assertEqual(CGU.setIntegerByPath(n, '/', 1)[1], numpy.array(1, dtype='int32'))
        self.assertTrue(numpy.array_equal(CGU.setIntegerByPath(n, '/', 1, 2, 3)[1], numpy.array([1, 2, 3], dtype='int32')))
        self.assertEqual(CGU.setLongByPath(n, '/', 1)[1], numpy.array(1, dtype='int64'))
        self.assertTrue(numpy.array_equal(CGU.setLongByPath(n, '/', 1, 2, 3)[1], numpy.array([1, 2, 3], dtype='int64')))
        self.assertEqual(CGU.setFloatByPath(n, '/', 1)[1], numpy.array(1, dtype='float32'))
        self.assertTrue(numpy.array_equal(CGU.setFloatByPath(n, '/', 1, 2, 3)[1], numpy.array([1, 2, 3], dtype='float32')))
        self.assertEqual(CGU.setDoubleByPath(n, '/', 1)[1], numpy.array(1, dtype='float64'))
        self.assertTrue(numpy.array_equal(CGU.setDoubleByPath(n, '/', 1, 2, 3)[1], numpy.array([1, 2, 3], dtype='float64')))
        self.assertTrue(numpy.array_equal(CGU.setDoubleByPath(n, '/', range(10, 1010, 10))[1],
                                          numpy.array([range(10, 1010, 10)], dtype='float64')))

    def test_06NodeType(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        A = ['A', None, [], None]
        B = ['B', None, [], None]
        C = ['C', None, [], None]
        D = ['D', None, [], None]
        A[2] += [B]
        A[2] += [C]
        C[3] = 'UnknownType_t'
        self.assertFalse(CGU.checkNodeType(C))
        self.assertFalse(CGU.checkNodeType(C, [CGK.CGNSBase_ts, CGK.Zone_ts]))
        self.assertFalse(CGU.checkNodeType(C, CGK.CGNSBase_ts))
        C[3] = CGK.CGNSBase_ts
        self.assertTrue(CGU.checkNodeType(C))
        self.assertTrue(CGU.checkNodeType(C, [CGK.CGNSBase_ts, CGK.Zone_ts]))
        self.assertTrue(CGU.checkNodeType(C, [CGK.Zone_ts, CGK.CGNSBase_ts,
                                              CGK.Zone_ts]))
        self.assertTrue(CGU.checkNodeType(C, CGK.CGNSBase_ts))

    def test_07NodeCompliance(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        import numpy
        n = CGU.nodeCreate('Base', numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        z = CGU.nodeCreate('ReferenceState', None, [], CGK.ReferenceState_ts, parent=n)
        self.assertTrue(CGU.checkNodeCompliant(n, None))
        self.assertTrue(CGU.checkNodeCompliant(z, None))
        self.assertTrue(CGU.checkNodeCompliant(z, n, dienow=True))
        self.assertFalse(CGU.checkNodeCompliant(None, n))

    def test_08NodeRetrieval(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        self.genTree()
        p1 = '/CGNSTree/{Base#1}/{Zone-B}'
        self.assertEqual(CGU.getNodeByPath(self.T, p1)[0], '{Zone-B}')
        p2 = '/CGNSTree/.///{Base#1}/{Zone-B}/../{Zone-A}'
        self.assertEqual(CGU.getNodeByPath(self.T, p2)[0], '{Zone-A}')
        p3 = '/{Base#1}/{Zone-A}/ZoneBC'
        n3 = CGU.getNodeByPath(self.T, p3)
        self.assertEqual(n3[0], 'ZoneBC')
        self.assertEqual(CGU.getPathFromNode(self.T, n3), p3)
        c1 = '{BC-2}'
        n31 = CGU.getNodeByPath(n3, c1)
        self.assertEqual(n31[0], '{BC-2}')
        self.assertEqual(CGU.getPathFromNode(n3, n31), "/" + c1)
        c2 = './{BC-1}'
        self.assertEqual(CGU.getNodeByPath(n3, c2)[0], '{BC-1}')
        filter = '/.*/.*/Zone.*'
        v1 = ['/{Base#1}/{Zone-B}/ZoneBC', '/{Base#1}/{Zone-B}/ZoneGridConnectivity']
        self.assertEqual(CGU.getPathByNameFilter(self.T, filter)[3:5], v1)
        filter = '/.*/.*/.*/GridConnectivity.*'
        v2 = '/{Base#1}/{Zone-D2}/ZoneGridConnectivity/{CT-D2-C}'
        self.assertEqual(CGU.getPathByTypeFilter(self.T, filter)[-2], v2)
        t = CGK.CGNSBase_ts
        b = CGU.getAncestorByType(self.T, n3, t)
        self.assertEqual(b[3], t)
        self.assertIsNone(CGU.getAncestorByType(self.T, b, CGK.Zone_ts))
        t = CGK.Zone_ts
        self.assertEqual(CGU.getAncestorByType(b, n3, t)[3], t)
        for p in CGU.getPathsByTypeSet(self.T, [CGK.BC_ts]):
            node = CGU.getNodeByPath(self.T, p)
            self.assertEqual(p, CGU.getPathFromNode(self.T, node))
        res = CGU.getPathsByTypeOrNameList(self.T, ['Zaza'])
        self.assertEqual(res, [])

    def test_09NodeDelete(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        import numpy
        n = CGU.nodeCreate('Base', numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        r = CGU.nodeCreate('ReferenceState', None, [], CGK.ReferenceState_ts, parent=n)
        d = CGU.nodeCreate('Data', numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.hasChildName(r, 'Data'))
        CGU.nodeDelete(n, d)
        self.assertIsNone(CGU.hasChildName(r, 'Data'))
        d = CGU.nodeCreate('DataZ', numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.hasChildName(r, 'DataZ'))
        CGU.nodeDelete(r, d)
        self.assertIsNone(CGU.hasChildName(r, 'DataZ'))
        CGU.nodeDelete(r, d)
        self.assertIsNone(CGU.hasChildName(r, 'DataZ'))
        n = CGU.nodeCreate('Base', numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        r = CGU.nodeCreate('ReferenceState', None, [], CGK.ReferenceState_ts, parent=n)
        d = CGU.nodeCreate('Data', numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.getNodeByPath(n, '/Base/ReferenceState/Data'))
        CGU.nodeDelete(n, d)
        self.assertIsNone(CGU.hasChildName(r, 'Data'))
        n = CGU.nodeCreate('Base', numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        r = CGU.nodeCreate('ReferenceState', None, [], CGK.ReferenceState_ts, parent=n)
        d = CGU.nodeCreate('Data', numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.getNodeByPath(n, '/Base/ReferenceState/Data'))
        CGU.removeChildByName(r, 'Data')
        self.assertIsNone(CGU.hasChildName(r, 'Data'))

# ---
print('-' * 70 + '\nCGNS.PAT test suite')
suite = unittest.TestLoader().loadTestsFromTestCase(PATTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS.NAV - Python package for CFD General Notation System - NAVigater
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import CGNS.PAT.cgnsutils as CGU

from qtpy.QtCore import (Qt, QModelIndex, QPoint)
from qtpy.QtWidgets import (QStyledItemDelegate, QMenu)
from qtpy.QtGui import (QColor, QFont, QIcon, QPixmap)

from CGNS.NAV.Q7MergeWindow import Ui_Q7MergeWindow
from CGNS.NAV.wfingerprint import Q7Window

import CGNS.NAV.mtree as NMT

(DIFF_NX, DIFF_NA, DIFF_ND, DIFF_CQ, DIFF_CT, DIFF_CS, DIFF_CV) = range(7)
(MERGE_NX, MERGE_NA, MERGE_NB) = range(3)


# -----------------------------------------------------------------
class Q7MergeItemDelegate(QStyledItemDelegate):
    def __init__(self, owner, model, diag, merge):
        QStyledItemDelegate.__init__(self, owner)
        self._parent = owner
        self._model = model
        self._diag = diag
        self._merge = merge
        self.orange = QColor("#FFA500")
        self.red = QColor("#FF0000")
        self.green = QColor("#008000")
        self.gray = QColor("#C0C0C0")

    def paint(self, painter, option, index):
        idx = self._parent.modelIndex(index)
        col = idx.column()
        nnm = self._parent.modelData(index).sidsName()
        pth = CGU.getPathNoRoot(self._parent.modelData(index).sidsPath())
        if (self._merge is not None) and (pth in self._merge):
            if self._merge[pth] == MERGE_NA:
                self._parent.modelData(index).setUserStatePrivate('A')
            if self._merge[pth] == MERGE_NB:
                self._parent.modelData(index).setUserStatePrivate('B')
        if col == NMT.COLUMN_NAME:
            if nnm not in OCTXT._ReservedNames:
                option.font.setWeight(QFont.Bold)
            QStyledItemDelegate.paint(self, painter, option, index)
            option.font.setWeight(QFont.Light)
            if (self._diag is not None) and (pth in self._diag):
                color = self.gray
                if self._diag[pth] == DIFF_NX:
                    color = self.gray
                if self._diag[pth] == DIFF_NA:
                    color = self.green
                if self._diag[pth] == DIFF_ND:
                    color = self.red
                painter.save()
                painter.setBrush(Qt.NoBrush)
                painter.setPen(color)
                painter.drawRect(option.rect)
                painter.restore()
        elif col == NMT.COLUMN_VALUE:
            option.font.setFamily(OCTXT.Table_Family)
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CV:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        elif col == NMT.COLUMN_DATATYPE:
            option.font.setFamily(OCTXT.Table_Family)
            option.font.setPointSize(8)
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CQ:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        elif col == NMT.COLUMN_SHAPE:
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CS:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        elif col == NMT.COLUMN_SIDS:
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CT:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        else:
            QStyledItemDelegate.paint(self, painter, option, index)


# -----------------------------------------------------------------
class Q7Merge(Q7Window, Ui_Q7MergeWindow):
    def __init__(self, control, fgprint, diag):
        Q7Window.__init__(self, Q7Window.VIEW_DIFF, control, None, fgprint)
        self._depthExpanded = 0
        self._lastEntered = None
        self._fgprint = fgprint
        (ldiag, lmerge) = self.diagAnalysis(diag)
        self.treeview.expanded[QModelIndex].connect(self.expandNode)
        self.treeview.collapsed.connect(self.collapseNode)
        self.treeview.clicked[QModelIndex].connect(self.clickedNode)
        self.treeview.customContextMenuRequested[QPoint].connect(self.clickedNode)
        # QObject.connect(self.treeview,
        #                 SIGNAL("expanded(QModelIndex)"),
        #                 self.expandNode)
        # QObject.connect(self.treeview,
        #                 SIGNAL("collapsed()"),
        #                 self.collapseNode)
        # QObject.connect(self.treeview,
        #                 SIGNAL("clicked(QModelIndex)"),
        #                 self.clickedNode)
        # QObject.connect(self.treeview,
        #                 SIGNAL("customContextMenuRequested(QPoint)"),
        #                 self.clickedNode)
        self.bClose.clicked.connect(self.leave)
        self.bInfo.clicked.connect(self.infoTreeView)
        self.bZoomIn.clicked.connect(self.expandLevel)
        self.bZoomOut.clicked.connect(self.collapseLevel)
        self.bZoomAll.clicked.connect(self.expandMinMax)
        self.bSaveDiff.clicked.connect(self.savediff)
        self.bSelectA.clicked.connect(self.showSelected)
        self.bSelectB.clicked.connect(self.showSelected)
        self.bSelectOrderSwap.clicked.connect(self.swapSelected)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.popupmenu = QMenu()
        self.proxyA = self._fgprint.model
        self.treeview.setModel(self.proxyA)
        self.treeview.setItemDelegate(Q7MergeItemDelegate(self.treeview,
                                                          self._fgprint.model,
                                                          ldiag, lmerge))
        self.treeview.setControlWindow(self, self._fgprint.model)
        self.treeview.hideColumn(NMT.COLUMN_FLAG_LINK)
        self.treeview.hideColumn(NMT.COLUMN_FLAG_CHECK)
        self.treeview.hideColumn(NMT.COLUMN_FLAG_SELECT)
        self._fgprint.model.addA = True
        self._fgprint.model.addB = True
        self._A = QIcon(QPixmap(":/images/icons/user-A.png"))
        self._B = QIcon(QPixmap(":/images/icons/user-B.png"))
        self._order = 0  # A first

    def swapSelected(self):
        if not self._order:
            self.bSelectA.setIcon(self._B)
            self.bSelectB.setIcon(self._A)
            self._order = 1
        else:
            self.bSelectA.setIcon(self._A)
            self.bSelectB.setIcon(self._B)
            self._order = 0

    def showSelected(self):
        self._fgprint.model.addA = False
        self._fgprint.model.addB = False
        if self.bSelectA.isChecked():
            self._fgprint.model.addA = True
        if self.bSelectB.isChecked():
            self._fgprint.model.addB = True

    def diagAnalysis(self, diag):
        ldiag = {}
        lmerge = {}
        for k in diag:
            ldiag[k] = DIFF_NX
            for d in diag[k]:
                if d[0] == 'NA':
                    ldiag[d[1]] = DIFF_NA
                    lmerge[d[1]] = MERGE_NB
                if d[0] == 'ND':
                    ldiag[d[1]] = DIFF_ND
                    lmerge[d[1]] = MERGE_NA
                if d[0] in ['CT']:
                    ldiag[k] = DIFF_CT
                    lmerge[k] = MERGE_NA
                if d[0] in ['C3', 'C1', 'C2']:
                    ldiag[k] = DIFF_CQ
                    lmerge[k] = MERGE_NA
                if d[0] in ['C6', 'C7']:
                    ldiag[k] = DIFF_CV
                    lmerge[k] = MERGE_NA
                if d[0] in ['C4', 'C5']:
                    ldiag[k] = DIFF_CS
                    lmerge[k] = MERGE_NA
        return (ldiag, lmerge)

    def model(self):
        return self._fgprint.model

    def modelIndex(self, idx):
        if not idx.isValid():
            return -1
        midx = idx
        # if (idx.model() != self.treeview.M()):
        #    midx=self.treeview.model().mapToSource(idx)
        return midx

    def modelData(self, idx):
        if not idx.isValid():
            return None
        return self.modelIndex(idx).internalPointer()

    def infoTreeView(self):
        self._control.helpWindow('Tree')

    def savediff(self):
        pass

    def expandMinMax(self):
        if self._depthExpanded == self._fgprint.depth - 2:
            self._depthExpanded = -1
            self.treeview.collapseAll()
        else:
            self._depthExpanded = self._fgprint.depth - 2
            self.treeview.expandAll()
        self.resizeAll()

    def expandLevel(self):
        if self._depthExpanded < self._fgprint.depth - 2:
            self._depthExpanded += 1
        self.treeview.expandToDepth(self._depthExpanded)
        self.resizeAll()

    def collapseLevel(self):
        if self._depthExpanded != -1:
            self._depthExpanded -= 1
        if self._depthExpanded == -1:
            self.treeview.collapseAll()
        else:
            self.treeview.expandToDepth(self._depthExpanded)
        self.resizeAll()

    def updateStatus(self, node):
        return
        self.lineEdit.clear()
        self.lineEdit.insert(node.sidsPath())

    def updateMenu(self, nodeidxs):
        return
        nodeidx = self.modelIndex(nodeidxs)
        if not nodeidx.isValid:
            return False
        if nodeidx.internalPointer() is None:
            return False
        if nodeidx.internalPointer().sidsPath() == '/CGNSTree':
            return False
        self.setLastEntered(nodeidxs)
        if nodeidx != -1:
            node = nodeidx.internalPointer()
            lknode = not node.sidsIsLink()
            lznode = node.hasLazyLoad()
            actlist = (
                ("About %s" % node.sidsType(), self.aboutSIDS, None, False),
                None,
                ("Mark/unmark node", self.marknode, 'Space', False),
                ("Add new child node", self.newnodechild, 'Ctrl+A', False),
                ("Add new brother node", self.newnodebrother, 'Ctrl+Z', False),
                None,
                ("Open form", self.popform, 'Ctrl+F', False),
                ("Open view", self.openSubTree, 'Ctrl+W', False),
                ("Open view on linked-to file", self.openLkTree, 'Ctrl+O', lknode),
                None,
                ("Load node data in memory", self.dataLoad, 'Ctrl+L', not lznode),
                ("Release memory node data", self.dataRelease, 'Ctrl+R', lznode),
                None,
                ("Copy", self.mcopy, 'Ctrl+C', False),
                ("Cut", self.mcut, 'Ctrl+X', False),
                ("Paste as brother", self.mpasteasbrother, 'Ctrl+V', False),
                ("Paste as child", self.mpasteaschild, 'Ctrl+Y', False),
                None,
                ("Cut all selected", self.mcutselected, 'Ctrl+Shift+X', False),
                ("Paste as brother for each selected",
                 self.mpasteasbrotherselected, 'Ctrl+Shift+V', False),
                ("Paste as child for each selected",
                 self.mpasteaschildselected, 'Ctrl+Shift+Y', False),
                ("Load nodes data in memory for each selected",
                 self.dataLoadSelected, 'Ctrl+Shift+L', False),
                ("Release memory node data for each selected",
                 self.dataReleaseSelected, 'Ctrl+Shift+R', False),
            )
            self.popupmenu.clear()
            self.popupmenu.setTitle('Node menu')
            for aparam in actlist:
                if aparam is None:
                    self.popupmenu.addSeparator()
                else:
                    a = QAction(aparam[0], self)
                    a.triggered.connect(aparam[1])
                    if aparam[2] is not None:
                        a.setShortcut(aparam[2])
                    self.popupmenu.addAction(a)
                    a.setDisabled(aparam[3])
            return True

    def setLastEntered(self, nix=None):
        self._lastEntered = None

    def getLastEntered(self):
        return self._lastEntered

    def clearLastEntered(self):
        self._lastEntered = None
        self.treeview.selectionModel().clearSelection()
        return None

    def clickedNode(self, index):
        pass

    def expandNode(self, *args):
        self.resizeAll()

    def collapseNode(self, *args):
        pass

    def resizeAll(self):
        for n in range(NMT.COLUMN_LAST + 1):
            self.treeview.resizeColumnToContents(n)

    def show(self):
        super(Q7Merge, self).show()

    def closeAlone(self):
        pass

    def leave(self):
        self.close()

# -----------------------------------------------------------------



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  ---------------------------------------------------------------------------
#
from __future__ import unicode_literals
#
import CGNS.PAT.cgnslib as C
import CGNS.PAT.cgnskeywords as K
import CGNS.PAT.cgnserrors as E
import numpy as N
#
#
data = C.newArea(None)
C.newUserDefinedData(data, '{UserDefinedData}')
C.newDescriptor(data, '{Descriptor}')
#
status = '11.3'
comment = 'Full SIDS with all optionals'
pattern = [data, status, comment]
#



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
"""
  pyCGNS - Python package for CFD General Notation System

  This packages provides some libs and tools dedicated to the CGNS standard.

  Modules are:

    * CGNS.MAP
    * CGNS.APP
    * CGNS.NAV
    * CGNS.DAT
    * CGNS.VAL
    * CGNS.PAT
    
"""
#
from CGNS.pyCGNSconfig import version as __version__
#
# --- last line



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  ---------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import division
from builtins import (str, bytes, range, dict)

import numpy

import CGNS.PAT.cgnserrors as CE
import CGNS.PAT.cgnskeywords as CK
import CGNS.PAT.cgnsutils as CU


# =============================================================================
# MLL-like calls
# - every call takes a parent as argument. If the parent is present, the new
#   sub-tree is inserted in the parent child list. In all cases the call
#   returns the created sub-tree
# - some function are not MLL based
# - function patterns
#   newXXX :   creates a new XXX_t type node
#   updateXXX: updates fields in the XXX_t node
#   checkXXX:  check if node is ok for SIDS and for CGNS/Python
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
def newCGNSTree(version=CK.CGNSLIBRARYVERSION):
    """
    Top CGNS/Python tree node creation, not a CGNS/SIDS type::

      T=newCGNSTree()

    :param float32 version: force the CGNSLibraryVersion
    :return: a new :ref:`XCGNSTree_t` node
    :Remarks:
      - You *should* keep the returned node in a variable or reference to it in
        any other way, this tree root is a Python object that would be garbagged
        if its reference count reaches zero.
      - The `CGNSTree` node is a CGNS/Python node which has no existence in a
        disk HDF5 file.
      - Default version is taken from CGNS.PAT.cgnskeywords
    :Children:
      - :py:func:`newCGNSBase`

    """
    return newCGNS(version)


def newCGNS(version=CK.CGNSLIBRARYVERSION):
    node = [CK.CGNSLibraryVersion_s,
            numpy.array([version], dtype=numpy.float32),
            [],
            CK.CGNSLibraryVersion_ts]
    badnode = [CK.CGNSTree_s, None, [node], CK.CGNSTree_ts]
    return badnode


# -----------------------------------------------------------------------------
def newCGNSBase(tree, name, ncell, nphys):
    """
    *CGNSBase* node creation, the top node for topological contents::

      # The base is put in the `T` children list
      T=newCGNSTree()
      newBase(T,'Box-1',3,3)

      # No parent, you should fetch the new node using a variable
      B=newCGNSBase(None,'Box-2',3,3)

      # using tuple de-ref
      dims=(3,3)
      B=newCGNSBase(None,'Box-3',*dims)


    :arg CGNS/Python node tree: the parent node (`<node>` or `None`)
    :arg str name: base name
    :arg int ncell: cell dimensions
    :arg int nphys: physical dimensions
    :return: a new :ref:`XCGNSBase_t` node
    :Children:
      - :py:func:`newZone`
      - :py:func:`newFamily`
      - :py:func:`newReferenceState`
      - :py:func:`newFlowEquationSet`
      - :py:func:`newConvergenceHistory`
    :See:
      - :rsids:`SIDS CGNSBase <cgnsbase#CGNSBase>`
    """
    return newBase(tree, name, ncell, nphys)


def newBase(tree, name, ncell, nphys):
    if ncell not in [1, 2, 3]:
        raise CE.cgnsException(10, name)
    if nphys not in [1, 2, 3]:
        raise CE.cgnsException(11, name)
    if nphys < ncell:
        raise CE.cgnsException(12, name)
    if (tree is not None) and (not CU.checkNode(tree)):
        raise CE.cgnsException(6, name)
    if (tree is not None) and (tree[0] == CK.CGNSTree_s):
        parent = tree[2]
    else:
        parent = tree
    CU.checkDuplicatedName(["<root node>", None, parent], name)
    node = CU.newNode(name,
                      numpy.array([ncell, nphys], dtype=numpy.int32, order='F'),
                      [], CK.CGNSBase_ts)
    if parent is not None:
        parent.append(node)
    return node


def numberOfBases(tree):
    return len(CU.hasChildType(tree, CK.CGNSBase_ts))


def readBase(tree, name):
    b = CU.hasChildName(tree, name)
    if b is None:
        raise CE.cgnsException(21, name)
    if b[3] != CK.CGNSBase_ts:
        raise CE.cgnsException(20, (CK.CGNSBase_ts, name))
    return b[0], b[1]


def updateBase(tree, name=None, ncell=None, nphys=None):
    if ncell not in [1, 2, 3]:
        raise CE.cgnsException(10, name)
    if nphys not in [1, 2, 3]:
        raise CE.cgnsException(11, name)
    if nphys < ncell:
        raise CE.cgnsException(12, name)

    if tree:
        CU.checkNode(tree)

    if tree[3] != CK.CGNSBase_ts:
        raise CE.cgnsException(20, (CK.CGNSBase_ts, name))
    if name is not None:
        tree[0] = name
    if (ncell is not None) and (nphys is not None) and tree:
        tree[1] = numpy.array([ncell, nphys], dtype=numpy.int32, order='F')
    else:
        raise CE.cgnsException(12)


# -----------------------------------------------------------------------------
def newZone(parent, name, zsize=None,
            ztype=CK.Structured_s,
            family=''):
    """
    *Zone* node creation, the sub-tree defining a topological domain::

      s=NPY.array([[10],[2],[0]],dtype=NPY.int32)

      T=newCGNSTree()
      B=newBase(T,'Box-1',3,3)
      Z=newZone(B,name,s,CK.Unstructured_s,'Wing')

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: zone name
    :arg ndarray zsize: array of zone dimensions
    :arg str ztype: type of the zone
    :arg str family: main FamilyName of the zone
    :return: a new :ref:`XZone_t` node
    :Children:
      - :py:func:`newElements`
    :Remarks:
      - The zone size has dimensions [IndexDimensions][3]
    :See:
      - :rsids:`SIDS Zone <cgnsbase#Zone>`

    """
    if ztype not in CK.ZoneType_l:
        raise CE.cgnsException(206, ztype)
    if zsize is None:
        raise CE.cgnsException(300)
    CU.checkDuplicatedName(parent, name)
    CU.checkArray(zsize, dienow=True)
    znode = CU.newNode(name, zsize, [], CK.Zone_ts, parent)
    CU.newNode(CK.ZoneType_s, CU.setStringAsArray(ztype), [], CK.ZoneType_ts, znode)
    if family:
        CU.newNode(CK.FamilyName_s,
                   CU.setStringAsArray(family), [], CK.FamilyName_ts, znode)
    return znode


def numberOfZones(tree, basename):
    b = CU.hasChildName(tree, basename)
    if b is None:
        raise CE.cgnsException(21, basename)
    if b[3] != CK.CGNSBase_ts:
        raise CE.cgnsException(20, (CK.CGNSBase_ts, basename))
    return len(CU.hasChildType(b, CK.Zone_ts))


def listZones(tree, basename=None):
    if basename is None:
        filterpath = [CK.CGNSTree_ts, CK.CGNSBase_ts, CK.Zone_ts]
    else:
        filterpath = [CK.CGNSTree_ts, basename, CK.Zone_ts]
    return CU.getPathsByTypeOrNameList(tree, filterpath)


def readZone(tree, basename, zonename, gtype=None):
    b = CU.hasChildName(tree, basename)
    if b is None:
        raise CE.cgnsException(21, basename)
    if b[3] != CK.CGNSBase_ts:
        raise CE.cgnsException(20, (CK.CGNSBase_ts, basename))
    z = CU.hasChildName(b, zonename)
    if z is None:
        raise CE.cgnsException(21, zonename)
    if z[3] != CK.Zone_ts:
        raise CE.cgnsException(20, (CK.Zone_ts, zonename))
    if gtype:
        zt = CU.hasChildName(z, CK.ZoneType_s)
        if zt is None:
            raise CE.cgnsException(21, CK.ZoneType_s)
        return z[0], z[1], zt[1]
    else:
        return z[0], z[1]


def typeOfZone(tree, basename, zonename):
    return readZone(tree, basename, zonename, 1)[2]


# -----------------------------------------------------------------------------
def newGridCoordinates(parent, name):
    """
    *GridCoordinates* node creation, container for coordinates::

      newGridCoordinates(zone,CK.GridCoordinates_s)
      newDataArray(gc,CK.CoordinateX_s)
      newDataArray(gc,CK.CoordinateY_s)
      newDataArray(gc,CK.CoordinateZ_s)

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :return: a new :ref:`XGridCoordinates_t` node
    :Remarks:
      - Creates only the grid node, you have to use :py:func:`newDataArray` to
        add the actual coordinates as children arrays. You can also use the
        :py:func:`newCoordinates` function, it creates this *GridCoordinates*
        node if missing, at the first coordinate add.
    :See:
      - :rsids:`SIDS GridCoordinates <gridflow#GridCoordinates>`
    """
    node = CU.newNode(name, None, [], CK.GridCoordinates_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newCoordinates(parent, name=CK.GridCoordinates_s, value=None):
    """
    *GridCoordinates_t* and *DataArray_t* nodes creation::

      cx=newCoordinates(zone,CK.CoordinateX_s,x_array)
      cy=newCoordinates(zone,CK.CoordinateY_s,y_array)
      cz=newCoordinates(zone,CK.CoordinateZ_s,z_array)

      # the following function sequence performs the same action
      gc=newGridCoordinates(zone,CK.GridCoordinates_s)
      newDataArray(gc,CK.CoordinateX_s,x_array)
      newDataArray(gc,CK.CoordinateY_s,y_array)
      newDataArray(gc,CK.CoordinateZ_s,z_array)

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :arg ndarray value: the coordinates array to set
    :return: The returned node always is the :ref:`XDataArray_t` node.
    :Remarks:
      - Creates a new <node> representing a GridCoordinates_t sub-tree with
        the coordinate DataArray given as argument. This creates both the
        GridCoordinates_t with GridCoordinates name and DataArray_t with the
        argument name. Usually used to create the default grid.
        If the GridCoordinates_t with name GridCoordinates already exists then
        only the DataArray is created.
      - Array dims are not checked, should be consistent with zone dims
    :See:
      - :rsids:`SIDS GridCoordinates <gridflow#GridCoordinates>`
    """
    CU.checkDuplicatedName(parent, name)
    gnode = CU.hasChildName(parent, CK.GridCoordinates_s)
    if gnode is None:
        gnode = newGridCoordinates(parent, CK.GridCoordinates_s)
    node = newDataArray(gnode, name, value)
    return node


# -----------------------------------------------------------------------------
def newDataArray(parent, name, value=None):
    """
    *DataArray* node creation, the all purpose array node::

      import numpy as NPY

      da=newDataArray(dd,'{DataArray}',value=NPY.array(((1,3),(5,7)),dtype=NPY.float64))

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: array name
    :arg value: optional value used to set the new array
    :return: a new :ref:`XDataArray_t` node
    :See:
      - :rsids:`SIDS DataArray <data#DataArray>`

    """
    CU.checkDuplicatedName(parent, name)
    # code correction:  Add value type and fortran order
    # code correction:  Add a specific array for string type
    # code correction:  Modify array check
    if isinstance(value, int):
        vv = numpy.array([value], dtype=numpy.int64, order='F')
        CU.checkArray(vv)
    elif isinstance(value, float):
        vv = numpy.array([value], dtype=numpy.float64, order='F')
        CU.checkArray(vv)
    elif isinstance(value, str):
        vv = CU.setStringAsArray(value)
        CU.checkArrayChar(vv)
    else:
        vv = value
        if vv is not None:
            CU.checkArray(vv)

    node = CU.newNode(name, vv, [], CK.DataArray_ts, parent)
    return node


def numberOfDataArrays(parent):
    return len(CU.hasChildType(parent, CK.DataArray_ts))


def readDataArray(parent, name):
    n = CU.hasChildName(parent, name)
    if n is None:
        raise CE.cgnsException(21, name)
    if n[3] != CK.DataArray_ts:
        raise CE.cgnsException(20, (CK.DataArray_ts, name))
    return n[1]


# -----------------------------------------------------------------------------
def newDataClass(parent, value=CK.UserDefined_s):
    """
    *DataClass* node creation, sets the class of a data array::

      import CGNS.PAT.cgnskeywords as CK
      import numpy as NPY

      # N is an already existing CGNS/Python node
      dd=newDiscreteData(N,'{DiscreteData}')
      dc=newDataClass(dd,CK.DimensionalUnits_s)
      da=newDataArray(dd,'{DataArray}',value=NPY.array((1,),dtype=NPY.float64))

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str value: DataClass type to set
    :return: a new :ref:`XDataClass_t` node
    :Remarks:
      - The value argument is a
        :ref:`DataClass enumerate <pat_cgnskeywords>` enumerate.
    :See:
      - :rsids:`SIDS DataClass <build#DataClass>
    """
    CU.checkDuplicatedName(parent, CK.DataClass_s)
    node = CU.newNode(CK.DataClass_s, CU.setStringAsArray(value), [],
                      CK.DataClass_ts, parent)
    return checkDataClass(node)


def updateDataClass(node, value):
    CU.checkNode(node)
    if value is not None:
        node[1] = value
    return checkDataClass(node)


def checkDataClass(node, parent=None):
    CU.checkNode(node)
    CU.checkName(node[0])
    if node[0] != CK.DataClass_s:
        raise CE.cgnsException(26, node[0])
    if node[3] != CK.DataClass_ts:
        raise CE.cgnsException(27, node[3])
    if len(node[2]) != 0:
        raise CE.cgnsException(28, node[0])
    value = CU.getValue(node).tostring().decode('ascii', 'strict')
    if value not in CK.DataClass_l:
        raise CE.cgnsException(207, value)
    if parent is not None:
        CU.checkTypeList(parent, [CK.DataArray_ts, CK.CGNSBase_ts, CK.Zone_ts,
                                  CK.GridCoordinates_ts, CK.Axisymmetry_ts,
                                  CK.RotatingCoordinates_ts, CK.FlowSolution_ts,
                                  CK.Periodic_ts, CK.ZoneBC_ts, CK.BC_ts,
                                  CK.BCDataSet_ts,
                                  CK.BCData_ts, CK.FlowEquationSet_ts,
                                  CK.GasModel_ts,
                                  CK.ViscosityModel_ts,
                                  CK.ThermalConductivityModel_ts,
                                  CK.TurbulenceClosure_ts, CK.TurbulenceModel_ts,
                                  CK.ThermalRelaxationModel_ts,
                                  CK.ChemicalKineticsModel_ts,
                                  CK.EMElectricFieldModel_ts,
                                  CK.EMMagneticFieldModel_ts,
                                  CK.EMConductivityModel_ts,
                                  CK.BaseIterativeData_ts,
                                  CK.ZoneIterativeData_ts, CK.RigidGridMotion_ts,
                                  CK.ArbitraryGridMotion_ts, CK.ReferenceState_ts,
                                  CK.ConvergenceHistory_ts,
                                  CK.DiscreteData_ts, CK.IntegralData_ts,
                                  CK.UserDefinedData_ts, CK.Gravity_ts]
                         , CK.DataClass_s)
    return node


# -----------------------------------------------------------------------------
def newDimensionalUnits(parent, value=None):
    """
    *DimensionalUnits* node creation, sets the units of a data array::

      import CGNS.PAT.cgnskeywords as CK
      import numpy as NPY

      # N is an already existing CGNS/Python node
      dd=newDiscreteData(N,'{DiscreteData}')
      dc=newDataClass(dd,CK.DimensionalUnits_s)
      units=(CK.Gram_s,CK.Foot_s,CK.UserDefined_s,CK.Celcius_s,CK.Degree_s)
      du=newDimensionalUnits(dd,units)
      da=newDataArray(dd,'{DataArray}',value=NPY.array((1,),dtype=NPY.float64))

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg list(str) value: list of 5 units, order is significant
    :return: a new :ref:`XDimensionalUnits_t` node
    :Remarks:
      - order is MassUnit,LengthUnit,TimeUnit,
        TemperatureUnit,AngleUnit
    :See:
      - :rsids:`SIDS DimensionalUnits <build#DimensionalUnits>`

    """
    if value is None:
        value = [CK.Kilogram_s, CK.Meter_s,
                 CK.Second_s, CK.Kelvin_s, CK.Radian_s]
    if len(value) != 5:
        raise CE.cgnsException(202)
    CU.checkDuplicatedName(parent, CK.DimensionalUnits_s)
    # --- loop over values to find all required units
    vunit = [CK.Null_s, CK.Null_s, CK.Null_s, CK.Null_s, CK.Null_s]
    for v in value:
        if v not in CK.AllUnits_l:
            raise CE.cgnsException(203, v)
        if ((v in CK.MassUnits_l) and
                (v not in [CK.Null_s, CK.UserDefined_s])):
            if v in vunit:
                raise CE.cgnsException(204, v)
            else:
                vunit[0] = v
        if ((v in CK.LengthUnits_l) and
                (v not in [CK.Null_s, CK.UserDefined_s])):
            if v in vunit:
                raise CE.cgnsException(204, v)
            else:
                vunit[1] = v
        if ((v in CK.TimeUnits_l) and
                (v not in [CK.Null_s, CK.UserDefined_s])):
            if v in vunit:
                raise CE.cgnsException(204, v)
            else:
                vunit[2] = v
        if ((v in CK.TemperatureUnits_l) and
                (v not in [CK.Null_s, CK.UserDefined_s])):
            if v in vunit:
                raise CE.cgnsException(204, v)
            else:
                vunit[3] = v
        if ((v in CK.AngleUnits_l) and
                (v not in [CK.Null_s, CK.UserDefined_s])):
            if v in vunit:
                raise CE.cgnsException(204, v)
            else:
                vunit[4] = v
    node = CU.newNode(CK.DimensionalUnits_s, CU.concatenateForArrayChar(vunit), [],
                      CK.DimensionalUnits_ts, parent)
    CU.newNode(CK.AdditionalUnits_s,
               CU.concatenateForArrayChar([CK.Null_s, CK.Null_s, CK.Null_s]),
               [],
               CK.AdditionalUnits_ts, node)
    return node


# -----------------------------------------------------------------------------
def newDimensionalExponents(parent,
                            MassExponent=0, LengthExponent=0, TimeExponent=0,
                            TemperatureExponent=0, AngleExponent=0):
    """
    *DimensionalExponents* node creation, sets the units exponents of an array::

      import CGNS.PAT.cgnskeywords as CK
      import numpy as NPY

      # N is an already existing CGNS/Python node
      dd=newDiscreteData(N,'{DiscreteData}')
      dc=newDataClass(dd,CK.DimensionalUnits_s)
      units=(CK.Gram_s,CK.Foot_s,CK.UserDefined_s,CK.Celcius_s,CK.Degree_s)
      du=newDimensionalUnits(dd,units)
      exps=(1,-1,-2,0,0)
      du=newDimensionalExponents(dd,exps)
      da=newDataArray(dd,'{DataArray}',value=NPY.array((1,),dtype=NPY.float64))

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg float MassExponent: exponent for mass
    :arg float LengthExponent: exponent for length
    :arg float TimeExponent: exponent for time
    :arg float TemperatureExponent: exponent for temperature
    :arg float AngleExponent: exponent for angle

    :return: a new :ref:`XDimensionalExponents_t` node
    :Remarks:
      - order is MassExponent,LengthExponent,TimeExponent,
        TemperatureExponent,AngleExponent
      - values are forced to be double floats
    :See:
      - :rsids:`SIDS DimensionalExponents <build#DimensionalExponents>`

    """
    CU.checkDuplicatedName(parent, CK.DimensionalExponents_s)
    node = CU.newNode(CK.DimensionalExponents_s,
                      numpy.array([MassExponent,
                                   LengthExponent,
                                   TimeExponent,
                                   TemperatureExponent,
                                   AngleExponent], dtype=numpy.float64, order='F'),
                      [], CK.DimensionalExponents_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newDataConversion(parent, ConversionScale=1.0, ConversionOffset=1.0):
    """
    *DataConversion* node creation, sets the conversion factors for an array::

      import CGNS.PAT.cgnskeywords as CK
      import numpy as NPY

      # N is an already existing CGNS/Python node
      dd=newDiscreteData(N,'{DiscreteData}')
      dc=newDataClass(dd,CK.DimensionalUnits_s)
      units=(CK.Gram_s,CK.Foot_s,CK.UserDefined_s,CK.Celcius_s,CK.Degree_s)
      du=newDimensionalUnits(dd,units)
      exps=(1,-1,-2,0,0)
      du=newDimensionalExponents(dd,exps)
      ds=newDataConversion(dd,2.0,0.0)
      da=newDataArray(dd,'{DataArray}',value=NPY.array((1,),dtype=NPY.float64))

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg float ConversionScale: scale of the conversion to apply
    :arg float ConversionOffset: offset of the conversion to apply
    :return: a new :ref:`XDataConversion_t` node
    :Remarks:
      - values are forced to be double floats
    :See:
      - :rsids:`SIDS DataConversion <data#DataConversion>`

    """
    CU.checkDuplicatedName(parent, CK.DataConversion_s)
    node = CU.newNode(CK.DataConversion_s,
                      numpy.array([ConversionScale, ConversionOffset],
                                  dtype=numpy.float64, order='F'),
                      [], CK.DataConversion_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newDescriptor(parent, name, value=''):
    """
    *Descriptor* node creation, to contain user-defined textual contents::

       txt=newDescriptor(parent,'CommandLine','python -c import elsA.CGNS')

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :arg str value: text to enter (python string)
    :return: a new :ref:`XDescriptor_t` node
    :See:
      - :rsids:`SIDS Descriptor <build#Descriptor>`

    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, CU.setStringAsArray(value), [], CK.Descriptor_ts, parent)
    return checkDescriptor(node)


def checkDescriptor(node, parent=None):
    CU.checkNode(node)
    CU.checkName(node[0])
    if node[3] != CK.Descriptor_ts:
        raise CE.cgnsException(27, node[3])
    if len(node[2]) != 0:
        raise CE.cgnsException(28, node[0])
    value = CU.getValue(node)
    if CU.getValueType(value) != CK.Character_s:
        raise CE.cgnsException(110, node[0])
    if parent is not None:
        CU.checkTypeList(parent, [CK.DataArray_ts, CK.CGNSBase_ts, CK.Zone_ts,
                                  CK.GridCoordinates_ts, CK.Elements_ts,
                                  CK.Axisymmetry_ts,
                                  CK.RotatingCoordinates_ts, CK.FlowSolution_ts,
                                  CK.ZoneGridConnectivity_ts,
                                  CK.GridConnectivity1to1_ts,
                                  CK.GridConnectivity_ts,
                                  CK.GridConnectivityProperty_ts,
                                  CK.AverageInterface_ts, CK.OversetHoles_ts,
                                  CK.Periodic_ts, CK.ZoneBC_ts, CK.BC_ts,
                                  CK.BCDataSet_ts,
                                  CK.BCData_ts, CK.FlowEquationSet_ts,
                                  CK.GasModel_ts,
                                  CK.BCProperty_ts, CK.WallFunction_ts,
                                  CK.Area_ts,
                                  CK.GoverningEquations_ts,
                                  CK.ViscosityModel_ts,
                                  CK.ThermalConductivityModel_ts,
                                  CK.TurbulenceClosure_ts, CK.TurbulenceModel_ts,
                                  CK.ThermalRelaxationModel_ts,
                                  CK.ChemicalKineticsModel_ts,
                                  CK.EMElectricFieldModel_ts,
                                  CK.EMMagneticFieldModel_ts,
                                  CK.EMConductivityModel_ts,
                                  CK.BaseIterativeData_ts,
                                  CK.ZoneIterativeData_ts, CK.RigidGridMotion_ts,
                                  CK.ArbitraryGridMotion_ts, CK.ReferenceState_ts,
                                  CK.ConvergenceHistory_ts,
                                  CK.DiscreteData_ts, CK.IntegralData_ts,
                                  CK.Family_ts, CK.GeometryReference_ts,
                                  CK.UserDefinedData_ts, CK.Gravity_ts]
                         , CK.DataClass_s)
    return node


# -----------------------------------------------------------------------------
def newGridLocation(parent, value=CK.CellCenter_s):
    """
    *GridLocation* node creation, set location of data value wrt grid::

      n=newGridLocation(parent,CK.Vertex_s)

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str value: GridLocation to set
    :return: a new :ref:`XGridLocation_t` node
    :Remarks:
      - Allowed values are in CK.GridLocation_l
    :See:
      - :rsids:`SIDS GridLocation <build#GridLocation>`
    """
    CU.checkDuplicatedName(parent, CK.GridLocation_s)
    if value not in CK.GridLocation_l:
        raise CE.cgnsException(200, value)
    node = CU.newNode(CK.GridLocation_s,
                      CU.setStringAsArray(value), [], CK.GridLocation_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newIndexArray(parent, name, value=None):
    """
    *IndexArray* node creation, integer array for indexing purpose::

      import numpy as NPY

      ix=newIndexArray(parent,'GlobalIndex',NPY.array((3,4,5)))

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :arg ndarray value: array to set
    :return: a new :ref:`XIndexArray_t` node
    :Remarks:
      - Values are *not* forced of any type, user has to use I4 or I8
      - Array dims are not checked
    :See:
      - :rsids:`SIDS IndexArray <build#IndexArray>`

    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, value, [], CK.IndexArray_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newPointList(parent, name=CK.PointList_s, value=None):
    """
    *PointList* node creation, integer array for indexing purpose::

      import numpy as NPY

      ix=newPointList(parent,'FacesList',NPY.array((3,4,5,9,15)))

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :arg ndarray value: array to set
    :return: a new :ref:`XPointList_t` node
    :Remarks:
      - Values are *not* forced of any type, user has to use I4 or I8
      - Array dims are not checked
      - Not a SIDS type
    :See:
      - :rsids:`SIDS IndexArray <build#IndexArray>`

    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, value, [], CK.IndexArray_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newPointRange(parent, name=CK.PointRange_s, value=None):
    """
    *PointRange* node creation, integer array for Structured indexing purpose::

      import numpy as NPY

      minmax=NPY.array([[1,13],[1,6],[1,1]],order='F')
      ix=newPointRange(parent,value=minmax)

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :arg ndarray value: array to set
    :return: a new :ref:`XPointRange_t` node
    :Remarks:
      - Values are *not* forced of any type, user has to use I4 or I8
      - Array dims are not checked
      - Array index has to be in Fortran order
      - not a SIDS type
    :See:
      - :rsids:`SIDS IndexRange <build#IndexRange>`

    """
    if value is None:
        value = []
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, value, [], CK.IndexRange_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newRind(parent, value):
    """
    *Rind* node creation, indicates extra *ghost* cells around the grid::

      rind=NPY.array([[1,13],[1,6],[1,1]],order='F')
      newRind(solution,rind)

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg ndarray value: array to set
    :return: a new :ref:`XRind_t` node
    :Remarks:
      - For structured grids, order is imin,imax,jmin,jmax,kmin,kmax
      - For unstructured grids, order is number of rind points before and after
      - Values are *not* forced of any type, user has to use I4 or I8
      - Array dims are not checked
      - Array index has to be in Fortran order
    :See:
      - :rsids:`SIDS Rind <build#Rind>`

    """
    CU.checkDuplicatedName(parent, CK.Rind_s)
    # check value wrt base dims
    node = CU.newNode(CK.Rind_s, value, [], CK.Rind_ts, parent)
    return node


# ----------------------------------------------------------------------------
def newSimulationType(parent, stype=CK.NonTimeAccurate_s):
    """
    *SimulationType* node creation, set the TimeAccurate type::

      newSimulationType(base,CK.TimeAccurate_s)

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str stype: enumerate from CK.SimulationType_l
    :return: a new :ref:`XSimulationType_t` node
    :See:
      - :rsids:`SIDS SimulationType <cgnsbase#CGNSBase>`

    """
    if parent:
        CU.checkNode(parent)
    CU.checkDuplicatedName(parent, CK.SimulationType_s)
    CU.checkType(parent, CK.CGNSBase_ts, CK.SimulationType_s)
    if stype not in CK.SimulationType_l:
        raise CE.cgnsException(205, stype)
    node = CU.newNode(CK.SimulationType_s,
                      CU.setStringAsArray(stype), [], CK.SimulationType_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newOrdinal(parent, value=0):
    """
    *Ordinal* node creation, an informative integer value::

      newOrdinal(node,4)

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg int value: arbtrary integer
    :return: a new :ref:`XOrdinal_t` node
    :See:
      - :rsids:`SIDS Ordinal <cgnsbase#Zone>`

    """
    CU.checkDuplicatedName(parent, CK.Ordinal_s)
    node = CU.newNode(CK.Ordinal_s,
                      numpy.array([value], dtype=numpy.int32), [], CK.Ordinal_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newDiscreteData(parent, name):
    """
    *DiscreteData* node creation, structural node for data::

      newDiscreteData(node,'Parameters')

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :return: a new :ref:`XDiscreteData_t` node
    :See:
      - :rsids:`SIDS DiscreteData <misc#DiscreteData>`

    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.DiscreteData_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newIntegralData(parent, name):
    """
    *IntegralData* node creation, structural node for data::

      newIntegralData(node,'Parameters')

    :arg CGNS/Python node parent: the parent node (`<node>` or `None`)
    :arg str name: new node name
    :return: a new :ref:`XIntegralData_t` node
    :See:
      - :rsids:`SIDS IntegralData <misc#IntegralData>`

    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.IntegralData_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newElements(parent, name,
                etype=CK.UserDefined_s,
                erange=None,
                econnectivity=None,
                estartoffset=None,
                eboundary=0
                ):
    """
    *Elements_t* node creation, indexing unstructured meshes::

     quads=newElements(None,'QUADS',CGK.QUAD_4,NPY.array([start,end]),quad_array)'

    - Args:
     * `parent`: the parent node (`<node>` or `None`)
     * `name`: element node name (`string`)
     * `etype`: the type of element (`string` or 'int')
     * `erange`: the first and last index of the connectivity (`numpy.ndarray`)
     * `econnectivity`: actual array of point connectivities (`numpy.ndarray`)
     * `estartoffset`: actual array of offsets for NGON/NFACES/MIXED connectivities (`numpy.ndarray`)
     * `eboundary`: number of boundary elements (`int`)

    - Return:
     * The new :ref:`XElements_t` node

    - Remarks:
     * If a parent is given, the new node is added to the parent children list.
     * The `elementsrange` *should* insure a unique and continuous index for
       all elements nodes in the same parent zone.
     * Element type can be set as `int` such as `CGK.QUAD_4` or 7, or
       as `string` such as `CGK.QUAD_4_s` or `"QUAD_4"`

    - Children:
     * :py:func:`newDescriptor`
    """
    CU.checkDuplicatedName(parent, name)
    if isinstance(etype, int):
        if etype not in CK.ElementType.values():
            raise CE.cgnsException(250, etype)
        etp = etype
    elif isinstance(etype, str) :
        etype=str(etype)
        if etype not in CK.ElementType_l:
            raise CE.cgnsException(250, etype)
        etp = CK.ElementType[etype]
    else:
        raise CE.cgnsException(250, etype)
    v = numpy.array([etp, eboundary], dtype=numpy.int32)
    enode = CU.newNode(name, v, [], CK.Elements_ts, parent)
    newPointRange(enode, CK.ElementRange_s, erange)
    newDataArray(enode, CK.ElementConnectivity_s, econnectivity)
    if etp in (CK.ElementType[CK.MIXED_s], CK.ElementType[CK.NGON_n_s],
            CK.ElementType[CK.NFACE_n_s]):
        newDataArray(enode, CK.ElementStartOffset_s, estartoffset)
    return enode

# -----------------------------------------------------------------------------
def newZoneBC(parent):
    return CU.newNode(CK.ZoneBC_s, None, [], CK.ZoneBC_ts, parent)


def newBC(parent, bname, brange=None,
          btype=CK.Null_s, bcType=CK.Null_s,
          family=CK.Null_s, pttype=CK.PointRange_s):
    if brange is None:
        brange = [[1, 1], [1, 1], [1, 1]]
    return newBoundary(parent, bname, brange, btype, family, pttype)


def newBoundary(parent, bname, brange,
                btype=CK.Null_s,
                family=None,
                pttype=CK.PointRange_s):
    """-BC node creation -BC

    'newNode:N='*newBoundary*'(parent:N,bname:S,brange:[*i],btype:S)'

    Returns a new <node> representing a BC_t sub-tree.
    If a parent is given, the new <node> is added to the parent children list.
    Parent should be Zone_t, returned node is parent.
    If the parent has already a child name ZoneBC then
    only the BC_t,IndexRange_t are created.
    chapter 9.3 Add IndexRange_t required
    """
    CU.checkDuplicatedName(parent, bname)
    zbnode = parent
    if ((zbnode is not None) and
            (zbnode[0] != CK.ZoneBC_s) and
            (zbnode[3] != CK.ZoneBC_ts)):
        zbnode = CU.newNode(CK.ZoneBC_s, None, [], CK.ZoneBC_ts, parent)
    bnode = CU.newNode(bname, CU.setStringAsArray(btype), [], CK.BC_ts, zbnode)
    if pttype == CK.PointRange_s:
        arange = numpy.array(brange, dtype=numpy.int32, order='F')
        CU.newNode(CK.PointRange_s, arange, [], CK.IndexRange_ts, bnode)
    else:
        arange = numpy.array(brange, dtype=numpy.int32, order='F')
        CU.newNode(CK.PointList_s, arange, [], CK.IndexArray_ts, bnode)
    if family:
        CU.newNode(CK.FamilyName_s, CU.setStringAsArray(family), [],
                   CK.FamilyName_ts, bnode)
    return bnode


# -----------------------------------------------------------------------------
def newBCDataSet(parent, name, valueType=CK.Null_s):
    """-BCDataSet node creation -BCDataSet

    'newNode:N='*newBCDataSet*'(parent:N,name:S,valueType:CK.BCTypeSimple)'

     If a parent is given, the new <node> is added to the parent children list.
     Returns a new <node> representing a BCDataSet_t sub-tree.
     chapter 9.4 Add node BCTypeSimple is required
    """
    node = CU.hasChildName(parent, name)
    if node is None:
        node = CU.newNode(name, None, [], CK.BCDataSet_ts, parent)
    if valueType not in CK.BCTypeSimple_l:
        raise CE.cgnsException(252, valueType)
    node[1] = CU.setStringAsArray(valueType)
    return node


# ---------------------------------------------------------------------------
def newBCData(parent, name=CK.NeumannData_s):
    """-BCData node creation -BCData

    'newNode:N='*newBCData*'(parent:N,name:S)'

     Returns a new <node> representing a BCData_t sub-tree.
     chapter 9.5
    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.BCData_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newBCProperty(parent, wallfunction=CK.Null_s, area=CK.Null_s):
    """-BCProperty node creation -BCProperty

    'newNode:N='*newBCProperty*'(parent:N)'

     Returns a new <node> representing a BCProperty_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter 9.6
    """
    CU.checkDuplicatedName(parent, CK.BCProperty_s)
    node = CU.newNode(CK.BCProperty_s, None, [], CK.BCProperty_ts, parent)
    wf = CU.newNode(CK.WallFunction_s, None, [], CK.WallFunction_ts, node)
    CU.newNode(CK.WallFunctionType_s, CU.setStringAsArray(wallfunction), [], CK.WallFunctionType_ts, wf)
    ar = newArea(node)
    return node

# -----------------------------------------------------------------------------
def newArea(parent, areatype=CK.UserDefined_s, regionname='{RegionName}',
            surfacearea=numpy.array([0.0])):
    ar = CU.newNode(CK.Area_s, None, [], CK.Area_ts, parent)
    CU.newNode(CK.AreaType_s, CU.setStringAsArray(areatype), [],
               CK.AreaType_ts, ar)
    CU.newNode(CK.RegionName_s, CU.setStringAsArray(regionname), [],
               CK.DataArray_ts, ar)
    newDataArray(ar, CK.SurfaceArea_s, surfacearea)
    
# -----------------------------------------------------------------------------
def newAxisymmetry(parent,
                   refpoint=numpy.array([0.0, 0.0, 0.0]),
                   axisvector=numpy.array([0.0, 0.0, 0.0])):
    """-Axisymmetry node creation -Axisymmetry

    'newNode:N='*newAxisymmetry*'(parent:N,refpoint:A,axisvector:A)'

    refpoint,axisvector should be a real array.
    Returns a new <node> representing a CK.Axisymmetry_t sub-tree.
    chapter 7.5 Add DataArray AxisymmetryAxisVector,AxisymmetryReferencePoint are required
    """
    if parent:
        CU.checkNode(parent)
    CU.checkType(parent, CK.CGNSBase_ts, CK.Axisymmetry_s)
    CU.checkDuplicatedName(parent, CK.Axisymmetry_s)
    CU.checkArrayReal(refpoint)
    CU.checkArrayReal(axisvector)
    node = CU.newNode(CK.Axisymmetry_s, None, [], CK.Axisymmetry_ts, parent)
    n = CU.hasChildName(parent, CK.AxisymmetryReferencePoint_s)
    if n is None:
        newDataArray(node, CK.AxisymmetryReferencePoint_s, numpy.array(refpoint))
    n = CU.hasChildName(parent, CK.AxisymmetryAxisVector_s)
    if n is None:
        newDataArray(node, CK.AxisymmetryAxisVector_s, numpy.array(axisvector))
    return node


# -----------------------------------------------------------------------------
def newRotatingCoordinates(parent,
                           rotcenter=numpy.array([0.0, 0.0, 0.0]),
                           ratev=numpy.array([0.0, 0.0, 0.0])):
    """-RotatingCoordinates node creation -RotatingCoordinates

    'newNode:N='*newRotatingCoordinates*'(parent:N,rotcenter=A,ratev=A)'

     Returns a new <node> representing a RotatingCoordinates_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     rotcenter,ratev should be a real array.
     chapter  7.6 Add DataArray RotationRateVector,RotationCenter are required
    """
    if parent:
        CU.checkNode(parent)
    CU.checkTypeList(parent, [CK.CGNSBase_ts, CK.Zone_ts, CK.Family_ts],
                     CK.RotatingCoordinates_s)
    CU.checkDuplicatedName(parent, CK.RotatingCoordinates_s)
    CU.checkArrayReal(rotcenter)
    CU.checkArrayReal(ratev)
    node = CU.newNode(CK.RotatingCoordinates_s, None, [], CK.RotatingCoordinates_ts, parent)
    n = CU.hasChildName(node, CK.RotationCenter_s)
    if n is None:
        newDataArray(node, CK.RotationCenter_s, numpy.array(rotcenter))
    n = CU.hasChildName(node, CK.RotationRateVector_s)
    if n is None:
        newDataArray(node, CK.RotationRateVector_s, numpy.array(ratev))
    return node


# -----------------------------------------------------------------------------
def newFlowSolution(parent, name='{FlowSolution}', gridlocation=None):
    """-Solution node creation -Solution

    'newNode:N='*newSolution*'(parent:N,name:S,gridlocation:None)'

    Returns a new <node> representing a FlowSolution_t sub-tree.
    chapter 7.7
    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.FlowSolution_ts, parent)
    if gridlocation is None:
        newGridLocation(node)
    else:
        newGridLocation(node, gridlocation)
    return node


# -----------------------------------------------------------------------------
def newZoneGridConnectivity(parent, name=CK.ZoneGridConnectivity_s):
    """-GridConnectivity node creation -Grid

    'newNode:N='*newZoneGridConnectivity*'(parent:N,name:S)'

    Creates a ZoneGridConnectivity_t sub-tree
    This sub-node is returned.
    If a parent is given, the new <node> is added to the parent children list,
    the parent should be a Zone_t.
    chapter 8.1
    """
    CU.checkDuplicatedName(parent, name)
    cnode = CU.hasChildName(parent, CK.ZoneGridConnectivity_s)
    if cnode is None:
        cnode = CU.newNode(CK.ZoneGridConnectivity_s,
                           None, [], CK.ZoneGridConnectivity_ts, parent)
    return cnode


# -----------------------------------------------------------------------------
def newGridConnectivity1to1(parent, name, dname, window, dwindow, trans):
    """-GridConnectivity1to1 node creation -Grid

    'newNode:N='*newGridConnectivity1to1*'(parent:N,name:S,dname:S,window:[i*],dwindow:[i*],trans:[i*])'

    Creates a GridConnectivity1to1_t sub-tree.
    If a parent is given, the new <node> is added to the parent children list,
    the parent should be a Zone_t.
    The returned node is the GridConnectivity1to1_t
    chapter 8.2
    """
    zcnode = CU.newNode(name, CU.setStringAsArray(dname), [],
                        CK.GridConnectivity1to1_ts, parent)
    CU.newNode(CK.Transform_s, numpy.array(list(trans), dtype=numpy.int32), [],
               CK.Transform_ts2, zcnode)
    CU.newNode(CK.PointRange_s,
               numpy.array(window, dtype=numpy.int32, order='F'), [],
               CK.IndexRange_ts, zcnode)
    CU.newNode(CK.PointRangeDonor_s,
               numpy.array(dwindow, dtype=numpy.int32, order='F'), [],
               CK.IndexRange_ts, zcnode)
    return zcnode


# -----------------------------------------------------------------------------
def newGridConnectivity(parent, name, dname, ctype=CK.Overset_s):
    """-GridConnectivity node creation -Grid

    'newNode:N='*newGridConnectivity*'(parent:N,name:S,dname:S,ctype:S)'

    Creates a GridConnectivity sub-tree.
    If a parent is given, the new <node> is added to the parent children list,
    the parent should be a ZoneGridConnectivity_t.
    The returned node is the GridConnectivity_t
    chapter 8.4
    """
    zcnode = CU.newNode(name, CU.setStringAsArray(dname),
                        [], CK.GridConnectivity_ts, parent)
    newGridConnectivityType(zcnode, ctype)
    return zcnode


# -----------------------------------------------------------------------------
def newGridConnectivityType(parent, ctype=CK.Overset_s):
    """-GridConnectivityType node creation -Grid

    'newNode:N='*newGridConnectivityType*'(parent:N,ctype:S)'

    Creates a GridConnectivityType sub-tree.
    If a parent is given, the new <node> is added to the parent children list,
    the parent should be a GridConnectivity_t.
    The returned node is the GridConnectivityType_t
    chapter 8
    """
    zctnode = CU.newNode(CK.GridConnectivityType_s, CU.setStringAsArray(ctype),
                         [], CK.GridConnectivityType_ts, parent)
    return zctnode


# -----------------------------------------------------------------------------
def newGridConnectivityProperty(parent):
    """-GridConnectivityProperty node creation -GridConnectivityProperty

    'newNode:N='*newGridConnectivityProperty*'(parent:N)'

     Returns a new <node> representing a GridConnectivityProperty_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter 8.5
    """
    CU.checkDuplicatedName(parent, CK.GridConnectivityProperty_s)
    nodeType = CU.newNode(CK.GridConnectivityProperty_s, None, [],
                          CK.GridConnectivityProperty_ts, parent)
    return nodeType


def newPeriodic(parent,
                rotcenter=numpy.array([0.0, 0.0, 0.0]),
                ratev=numpy.array([0.0, 0.0, 0.0]),
                trans=numpy.array([0.0, 0.0, 0.0])):
    """-Periodic node creation -Periodic

    'newNode:N='*newPeriodic*'(parent:N,rotcenter=A,ratev=A,trans=A)'

     Returns a new <node> representing a Periodic_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name Periodic then
     only the RotationCenter,RotationAngle,Translation are created.
     rotcenter,ratev,trans should be a real array.
     chapter 8.5.1 Add DataArray RotationCenter,RotationAngle,Translation are required
    """
    if parent:
        CU.checkNode(parent)
    CU.checkArrayReal(rotcenter)
    CU.checkArrayReal(ratev)
    CU.checkArrayReal(trans)
    cnode = CU.hasChildName(parent, CK.Periodic_s)
    if cnode is None:
        cnode = CU.newNode(CK.Periodic_s, None, [], CK.Periodic_ts, parent)
    n = CU.hasChildName(cnode, CK.RotationCenter_s)
    if n is None:
        newDataArray(cnode, CK.RotationCenter_s, numpy.array(rotcenter))
    n = CU.hasChildName(cnode, CK.RotationAngle_s)
    if n is None:
        newDataArray(cnode, CK.RotationAngle_s, numpy.array(ratev))
    n = CU.hasChildName(cnode, CK.Translation_s)
    if n is None:
        newDataArray(cnode, CK.Translation_s, numpy.array(trans))
    return cnode


# -----------------------------------------------------------------------------
def newAverageInterface(parent, valueType=CK.Null_s):
    """-AverageInterface node creation -AverageInterface

    'newNode:N='*newAverageInterface*'(parent:N,valueType:CK.AverageInterfaceType)'

     Returns a new <node> representing a AverageInterface_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name AverageInterface then
     only the AverageInterfaceType is created.
     chapter 8.5.2
    """
    node = CU.hasChildName(parent, CK.AverageInterface_s)
    if node is None:
        node = CU.newNode(CK.AverageInterface_s, None, [],
                          CK.AverageInterface_ts, parent)
    if valueType not in CK.AverageInterfaceType_l:
        raise CE.cgnsException(253, valueType)
    CU.checkDuplicatedName(node, CK.AverageInterfaceType_s)
    # code correction: Modify valueType string into NPY string array
    CU.newNode(CK.AverageInterfaceType_s, CU.setStringAsArray(valueType), [],
               CK.AverageInterfaceType_ts, node)
    return node


# -----------------------------------------------------------------------------
def newOversetHoles(parent, name, hrange):
    """-OversetHoles node creation -OversetHoles

    'node:N='*newOversetHoles*'(parent:N,name:S,hrange:list)'

    Creates a OversetHoles_t sub-tree.
    the parent should be a Zone_t.
    If a parent is given, the new <node> is added to the parent children list.
    chapter 8.6 Add PointList or List( PointRange ) are required
    """
    cnode = CU.hasChildName(parent, CK.ZoneGridConnectivity_s)
    if cnode is None:
        cnode = CU.newNode(CK.ZoneGridConnectivity_s, None, [], CK.ZoneGridConnectivity_ts, parent)
    CU.checkDuplicatedName(cnode, name)
    node = CU.newNode(name, None, [], CK.OversetHoles_ts, cnode)
    # if(pname!=None and value!=None):
    # newPointList(node,pname,value)
    if hrange is not None:
        # code correction: Modify PointRange shape and order
        newPointRange(node, CK.PointRange_s, numpy.array(hrange, dtype=numpy.int32, order='F'))
        # newNode(CK.PointRange_s,NPY.array(list(hrange),NPY.int32),[],CK.IndexRange_ts,node)
    return node


# -----------------------------------------------------------------------------
def newFlowEquationSet(parent):
    """-FlowEquationSet node creation -FlowEquationSet

    'newNode:N='*newFlowEquationSet*'(parent:N)'

    If a parent is given, the new <node> is added to the parent children list.
     Returns a new <node> representing a CK.FlowEquationSet_t sub-tree.
     chapter 10.1
    """
    if parent:
        CU.checkNode(parent)
    CU.checkDuplicatedName(parent, CK.FlowEquationSet_s)
    CU.checkTypeList(parent, [CK.CGNSBase_ts, CK.Zone_ts], CK.FlowEquationSet_s)
    node = CU.newNode(CK.FlowEquationSet_s, None, [], CK.FlowEquationSet_ts, parent)
    return node


def newGoverningEquations(parent, valueType=CK.Euler_s):
    """-GoverningEquations node creation -GoverningEquations

    'newNode:N='*newGoverningEquations*'(parent:N,valueType:CK.GoverningEquationsType)'

     Returns a new <node> representing a CK.GoverningEquations_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name GoverningEquations then
     only the GoverningEquationsType is created.
     chapter  10.2 Add node GoverningEquationsType is required
    """
    node = CU.hasChildName(parent, CK.GoverningEquations_s)
    if node is None:
        node = CU.newNode(CK.GoverningEquations_s, None, [], CK.GoverningEquations_ts, parent)
    if valueType not in CK.GoverningEquationsType_l:
        raise CE.cgnsException(221, valueType)
    CU.checkDuplicatedName(parent, CK.GoverningEquationsType_s, )
    node[1] = CU.setStringAsArray(valueType)
    return node


# -----------------------------------------------------------------------------
def newGasModel(parent, valueType=CK.Ideal_s):
    """-GasModel node creation -GasModel

    'newNode:N='*newGasModel*'(parent:N,valueType:CK.GasModelType)'

     Returns a new <node> representing a CK.GasModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name GasModel then
     only the GasModelType is created.
     chapter 10.3 Add node GasModelType is required
    """
    node = CU.hasChildName(parent, CK.GasModel_s)
    if node is None:
        node = CU.newNode(CK.GasModel_s, None, [], CK.GasModel_ts, parent)
    if valueType not in CK.GasModelType_l:
        raise CE.cgnsException(224, valueType)
    CU.checkDuplicatedName(node, CK.GasModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newThermalConductivityModel(parent, valueType=CK.SutherlandLaw_s):
    """-ThermalConductivityModel node creation -ThermalConductivityModel

    'newNode:N='*newThermalConductivityModel*'(parent:N,valueType:CK.ThermalConductivityModelType)'

     Returns a new <node> representing a CK.ThermalConductivityModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name ThermalConductivityModel then
     only the ThermalConductivityModelType is created.
     chapter 10.5 Add node ThermalConductivityModelType is required
    """
    node = CU.hasChildName(parent, CK.ThermalConductivityModel_s)
    if node is None:
        node = CU.newNode(CK.ThermalConductivityModel_s, None, [],
                          CK.ThermalConductivityModel_ts, parent)
    if valueType not in CK.ThermalConductivityModelType_l:
        raise CE.cgnsException(227, valueType)
    CU.checkDuplicatedName(node, CK.ThermalConductivityModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newViscosityModel(parent, valueType=CK.SutherlandLaw_s):
    """-ViscosityModel node creation -ViscosityModel

    'newNode:N='*newViscosityModel*'(parent:N,valueType:CK.ViscosityModelType)'

     Returns a new <node> representing a CK.ViscosityModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name ViscosityModel then
     only the ViscosityModelType is created.
     chapter 10.4 Add node ViscosityModelType is (r)
    """
    node = CU.hasChildName(parent, CK.ViscosityModel_s)
    if node is None:
        node = CU.newNode(CK.ViscosityModel_s, None, [], CK.ViscosityModel_ts, parent)
    if valueType not in CK.ViscosityModelType_l:
        raise CE.cgnsException(230, valueType)
    CU.checkDuplicatedName(node, CK.ViscosityModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newTurbulenceClosure(parent, valueType=CK.Null_s):
    """-TurbulenceClosure node creation -TurbulenceClosure

    'newNode:N='*newTurbulenceClosure*'(parent:N,valueType:CK.TurbulenceClosureType)'
     Returns a new <node> representing a CK.TurbulenceClosure_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name TurbulenceClosure then
     only the ViscosityModelType is created.
     chapter 10.5 Add node TurbulenceClosureType is (r)
    """
    node = CU.hasChildName(parent, CK.TurbulenceClosure_s)
    if node is None:
        node = CU.newNode(CK.TurbulenceClosure_s, None, [], CK.TurbulenceClosure_ts, parent)
    if valueType not in CK.TurbulenceClosureType_l:
        raise CE.cgnsException(233, valueType)
    CU.checkDuplicatedName(node, CK.TurbulenceClosureType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newTurbulenceModel(parent, valueType=CK.OneEquation_SpalartAllmaras_s):
    """-TurbulenceModel node creation -TurbulenceModel

    'newNode:N='*newTurbulenceModel*'(parent:N,valueType:CK.TurbulenceModelType)'

     Returns a new <node> representing a CK.TurbulenceModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name TurbulenceModel then
     only the TurbulenceModelType is created.
     chapter 10.6.2 Add node TurbulenceModelType is (r)
    """
    node = CU.hasChildName(parent, CK.TurbulenceModel_s)
    if node is None:
        node = CU.newNode(CK.TurbulenceModel_s, None, [], CK.TurbulenceModel_ts, parent)
    if valueType not in CK.TurbulenceModelType_l:
        raise CE.cgnsException(236, valueType)
    CU.checkDuplicatedName(node, CK.TurbulenceModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newThermalRelaxationModel(parent, valueType=CK.Null_s):
    """-ThermalRelaxationModel node creation -ThermalRelaxationModel

    'newNode:N='*newThermalRelaxationModel*'(parent:N,valueType:CK.ThermalRelaxationModelType)'

     Returns a new <node> representing a CK.ThermalRelaxationModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name ThermalRelaxationModel then
     only the ThermalRelaxationModelType is created.
     chapter 10.7 Add node ThermalRelaxationModelType is (r)
    """
    node = CU.hasChildName(parent, CK.ThermalRelaxationModel_s)
    if node is None:
        node = CU.newNode(CK.ThermalRelaxationModel_s, None, [],
                          CK.ThermalRelaxationModel_ts, parent)
    if valueType not in CK.ThermalRelaxationModelType_l:
        raise CE.cgnsException(239, valueType)
    CU.checkDuplicatedName(node, CK.ThermalRelaxationModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newChemicalKineticsModel(parent, valueType=CK.Null_s):
    """-ChemicalKineticsModel node creation -ChemicalKineticsModel

    'newNode:N='*newChemicalKineticsModel*'(parent:N,valueType:CK.ChemicalKineticsModelType)'

     Returns a new <node> representing a CK.ChemicalKineticsModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name ChemicalKineticsModel then
     only the ChemicalKineticsModelType is created.
     chapter 10.8 Add node ChemicalKineticsModelType is (r)
    """
    node = CU.hasChildName(parent, CK.ChemicalKineticsModel_s)
    if node is None:
        node = CU.newNode(CK.ChemicalKineticsModel_s, None, [],
                          CK.ChemicalKineticsModel_ts, parent)
    if valueType not in CK.ChemicalKineticsModelType_l:
        raise CE.cgnsException(242, valueType)
    CU.checkDuplicatedName(node, CK.ChemicalKineticsModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newEMElectricFieldModel(parent, valueType=CK.Null_s):
    """-EMElectricFieldModel node creation -EMElectricFieldModel

    'newNode:N='*newEMElectricFieldModel*'(parent:N,valueType:CK.EMElectricFieldModelType)'

     Returns a new <node> representing a CK.EMElectricFieldModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
      If the parent has already a child name EMElectricFieldModel then
     only the EMElectricFieldModelType is created.
     chapter 10.9 Add node EMElectricFieldModelType is (r)
    """
    node = CU.hasChildName(parent, CK.EMElectricFieldModel_s)
    if node is None:
        node = CU.newNode(CK.EMElectricFieldModel_s, None, [],
                          CK.EMElectricFieldModel_ts, parent)
    if valueType not in CK.EMElectricFieldModelType_l:
        raise CE.cgnsException(245, valueType)
    CU.checkDuplicatedName(node, CK.EMElectricFieldModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newEMMagneticFieldModel(parent, valueType=CK.Null_s):
    """-EMMagneticFieldModel node creation -EMMagneticFieldModel

    'newNode:N='*newEMMagneticFieldModel*'(parent:N,valueType:CK.EMMagneticFieldModelType)'

     Returns a new <node> representing a CK.EMMagneticFieldModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name EMMagneticFieldModel_s then
     only the EMMagneticFieldModelType is created.
     chapter 10.9.2 Add node EMMagneticFieldModelType is (r)
    """
    node = CU.hasChildName(parent, CK.EMMagneticFieldModel_s)
    if node is None:
        node = CU.newNode(CK.EMMagneticFieldModel_s, None, [],
                          CK.EMMagneticFieldModel_ts, parent)
    if valueType not in CK.EMMagneticFieldModelType_l:
        raise CE.cgnsException(248, valueType)
    CU.checkDuplicatedName(node, CK.EMMagneticFieldModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


def newEMConductivityModel(parent, valueType=CK.Null_s):
    """-EMConductivityModel node creation -EMConductivityModel

    'newNode:N='*newEMConductivityModel*'(parent:N,valueType:CK.EMConductivityModelType)'

     Returns a new <node> representing a CK.EMConductivityModel_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name EMConductivityModel then
     only the EMConductivityModelType is created.
     chapter 10.9.3 Add node EMConductivityModelType is (r)
    """
    node = CU.hasChildName(parent, CK.EMConductivityModel_s)
    if node is None:
        node = CU.newNode(CK.EMConductivityModel_s, None, [],
                          CK.EMConductivityModel_ts, parent)
    if valueType not in CK.EMConductivityModelType_l:
        raise CE.cgnsException(218, valueType)
    CU.checkDuplicatedName(node, CK.EMConductivityModelType_s)
    node[1] = CU.setStringAsArray(valueType)
    return node


# -----------------------------------------------------------------------------
def newBaseIterativeData(parent, name, nsteps=0,
                         itype=CK.IterationValues_s):
    """-BaseIterativeData node creation -BaseIterativeData

     'newNode:N='*newBaseIterativeData*'(parent:N,name:S,nsteps:I,itype:E)'

     Returns a new <node> representing a BaseIterativeData_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter 11.1.1
     NumberOfSteps is required, TimeValues or IterationValues are required
    """

    if parent:
        CU.checkNode(parent)
    CU.checkDuplicatedName(parent, name)
    CU.checkType(parent, CK.CGNSBase_ts, CK.BaseIterativeData_ts)
    if not isinstance(nsteps, int) or (nsteps < 0):
        raise CE.cgnsException(209)
    asteps = numpy.arange(1, nsteps + 1, dtype=numpy.int32)
    node = CU.newNode(name, numpy.array([nsteps], dtype=numpy.int32), [],
                      CK.BaseIterativeData_ts, parent)
    if itype not in [CK.IterationValues_s, CK.TimeValues_s]:
        raise CE.cgnsException(210, (CK.IterationValues_s, CK.TimeValues_s))
    CU.newNode(itype, asteps, [], CK.DataArray_ts, node)
    return node


# -----------------------------------------------------------------------------
def newZoneIterativeData(parent, name):
    """-ZoneIterativeData node creation -ZoneIterativeData

    'newNode:N='*newZoneIterativeData*'(parent:N,name:S)'

     Returns a new <node> representing a ZoneIterativeData_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter  11.1.2
    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.ZoneIterativeData_ts, parent)
    return node


# ---------------------------------------------------------------------------
def newRigidGridMotion(parent, name,
                       valueType=CK.Null_s,
                       vector=numpy.array([0.0, 0.0, 0.0])):
    """-RigidGridMotion node creation -RigidGridMotion

    'newNode:N='*newRigidGridMotion*'(parent:N,name:S,valueType:CK.RigidGridMotionType,vector:A)'

    If a parent is given, the new <node> is added to the parent children list.
     Returns a new <node> representing a CK.RigidGridMotion_t sub-tree.
     If the parent has already a child name RigidGridMotion then
     only the RigidGridMotionType is created and OriginLocation is created
     chapter 11.2 Add Node RigidGridMotionType and add DataArray OriginLocation are the only required
    """
    if parent:
        CU.checkNode(parent)
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.RigidGridMotion_ts, parent)

    if valueType not in CK.RigidGridMotionType_l:
        raise CE.cgnsException(254, valueType)
    node[1] = CU.setStringAsArray(valueType)
    n = CU.hasChildName(parent, CK.OriginLocation_s)
    if n is None:
        newDataArray(node, CK.OriginLocation_s, numpy.array(vector))
    return node


# -----------------------------------------------------------------------------
def newReferenceState(parent, name=CK.ReferenceState_s):
    """-ReferenceState node creation -ReferenceState

    'newNode:N='*newReferenceState*'(parent:N,name:S)'

     Returns a new <node> representing a ReferenceState_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter  12.1  """
    if parent:
        CU.checkNode(parent)
    node = CU.hasChildName(parent, name)
    if node is None:
        CU.checkDuplicatedName(parent, name)
        node = CU.newNode(name, None, [], CK.ReferenceState_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newConvergenceHistory(parent, name=CK.GlobalConvergenceHistory_s,
                          iterations=0):
    """-ConvergenceHistory node creation -ConvergenceHistory

    'newNode:N='*newConvergenceHistory*'(parent:N,name:S,iterations:i)'

     Returns a new <node> representing a ConvergenceHistory_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter  12.3  """
    if name not in CK.ConvergenceHistory_l:
        raise CE.cgnsException(201, name)
    if parent:
        CU.checkNode(parent)
        CU.checkTypeList(parent, [CK.CGNSBase_ts, CK.Zone_ts], name)
    if name == CK.GlobalConvergenceHistory_s:
        CU.checkType(parent, CK.CGNSBase_ts, name)
    if name == CK.ZoneConvergenceHistory_s:
        CU.checkType(parent, CK.Zone_ts, name)
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, numpy.array([iterations], dtype=numpy.int32), [], CK.ConvergenceHistory_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newFamily(parent, name):
    """-Family node creation -Family

    'newNode:N='*newFamily*'(parent:N,name:S)'

     Returns a new <node> representing a Family_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter  12.6
    """
    if parent:
        CU.checkNode(parent)
    CU.checkType(parent, CK.CGNSBase_ts, name)
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.Family_ts, parent)
    return node


def newFamilyName(parent, family=None):
    # code correction: Modify family string into NPY string array
    return CU.newNode(CK.FamilyName_s, CU.setStringAsArray(family),
                      [], CK.FamilyName_ts, parent)

def newAdditionalFamilyName(parent, family=None):
    # code correction: Modify family string into NPY string array
    return CU.newNode(CK.FamilyName_s, CU.setStringAsArray(family),
                      [], CK.AdditionalFamilyName_ts, parent)


# -----------------------------------------------------------------------------
def newGeometryReference(parent, name='{GeometryReference}',
                         valueType=CK.UserDefined_s):
    """-GeometryReference node creation -GeometryReference

    'newNode:N='*newGeometryReference*'(parent:N,name:S,valueType:CK.GeometryFormat)'

     Returns a new <node> representing a CK.GeometryFormat_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name CK.GeometryReference then
     only the .GeometryFormat is created
     chapter  12.7 Add node CK.GeometryFormat_t is (r) and GeometryFile_t definition not find but is required (CAD file)
    """
    node = CU.hasChildName(parent, CK.GeometryReference_s)
    if node is None:
        node = CU.newNode(name, None, [], CK.GeometryReference_ts, parent)
    if valueType not in CK.GeometryFormat_l:
        raise CE.cgnsException(256, valueType)
    CU.checkDuplicatedName(node, CK.GeometryFormat_s)
    # code correction: Modify valueType string into NPY string array
    CU.newNode(CK.GeometryFormat_s, CU.setStringAsArray(valueType), [],
               CK.GeometryFormat_ts, node)
    return node


# -----------------------------------------------------------------------------
def newFamilyBC(parent, valueType=CK.UserDefined_s):
    """-FamilyBC node creation -FamilyBC

    'newNode:N='*newFamilyBC*'(parent:N,valueType:CK.BCTypeSimple/CK.BCTypeCompound)'

     Returns a new <node> representing a CK.FamilyBC_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     If the parent has already a child name FamilyBC then
     only the BCType is created
     chapter  12.8 Add node BCType is required
    """
    node = CU.hasChildName(parent, CK.FamilyBC_s)
    if (valueType not in CK.BCTypeSimple_l and
            valueType not in CK.BCTypeCompound_l):
        raise CE.cgnsException(257, valueType)
    if node is None:
        node = CU.newNode(CK.FamilyBC_s,
                          CU.setStringAsArray(valueType),
                          [],
                          CK.FamilyBC_ts,
                          parent)
    return node


# -----------------------------------------------------------------------------
def newArbitraryGridMotion(parent, name, valuetype=CK.Null_s):
    """
    Returns a **new node** representing a :ref:`XArbitraryGridMotionType_t`

    :param parent: CGNS/Python node
    :param name: String
    :param valuetype: String (``CGNS.PAT.cgnskeywords.ArbitraryGridMotionType``)


    If a *parent* is not ``None``, the **new node** is added to the parent
    children list. If the *parent* has already a child with
    name ``RigidGridMotion`` then only the ``RigidGridMotionType`` is created.

    """
    node = None
    if parent:
        node = CU.hasChildName(parent, name)
    if node is None:
        node = CU.newNode(name, None, [], CK.ArbitraryGridMotion_ts, parent)
    if valuetype not in CK.ArbitraryGridMotionType_l:
        raise CE.cgnsException(255, valuetype)
    CU.checkDuplicatedName(node, CK.ArbitraryGridMotionType_s)
    node[1] = CU.setStringAsArray(valuetype)
    return node


# -----------------------------------------------------------------------------
def newUserDefinedData(parent, name):
    """-UserDefinedData node creation -UserDefinedData

    'newNode:N='*newUserDefinedData*'(parent:N,name:S)'

     Returns a new <node> representing a UserDefinedData_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     chapter  12.9
    """
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, None, [], CK.UserDefinedData_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newGravity(parent, gvector=numpy.array([0.0, 0.0, 0.0])):
    """-Gravity node creation -Gravity

    'newNode:N='*newGravity*'(parent:N,gvector:A)'

     Returns a new <node> representing a Gravity_t sub-tree.
     If a parent is given, the new <node> is added to the parent children list.
     gvector should be a real array
     chapter  12.10 Add DataArray GravityVector is required
    """
    if parent:
        CU.checkNode(parent)
    CU.checkType(parent, CK.CGNSBase_ts, CK.Gravity_s)
    CU.checkDuplicatedName(parent, CK.Gravity_s)
    CU.checkArrayReal(gvector)
    node = CU.newNode(CK.Gravity_s, None, [], CK.Gravity_ts, parent)
    n = CU.hasChildName(parent, CK.GravityVector_s)
    if n is None:
        newDataArray(node, CK.GravityVector_s, numpy.array(gvector))
    return node


# -----------------------------------------------------------------------------
def newField(parent, name, value):
    CU.checkDuplicatedName(parent, name)
    node = newDataArray(parent, name, value)
    return node


# -----------------------------------------------------------------------------
def newModel(parent, name, label, value):
    CU.checkDuplicatedName(parent, name)
    node = CU.newNode(name, value, [], label, parent)
    return node


# -----------------------------------------------------------------------------
def newDiffusionModel(parent, value=None):
    # the diffusion_t doesn't exist. We use the cgnspatch file to keep
    # track of this...
    CU.checkDuplicatedName(parent, CK.DiffusionModel_s)
    node = CU.newNode(CK.DiffusionModel_s, value, [], CK.DiffusionModel_ts, parent)
    return node


# -----------------------------------------------------------------------------
# def newSection():
#  pass

# -----------------------------------------------------------------------------
def newParentElements(parent, value):
    CU.checkDuplicatedName(parent, CK.ParentElements_s)
    node = CU.newNode(CK.ParentElements_s, value, [], CK.DataArray_ts, parent)
    return node


# -----------------------------------------------------------------------------
def newParentElementsPosition(parent, value):
    CU.checkDuplicatedName(parent, CK.ParentElementsPosition_s)
    node = CU.newNode(CK.ParentElementsPosition_s, value, [], CK.DataArray_ts, parent)
    return node


# -----------------------------------------------------------------------------
# def newPart():
#  pass

# -----------------------------------------------------------------------------
def nextRange(previous, etype, earray):
    r = previous
    if previous is None:
        r = numpy.array([0, 0], dtype=numpy.int32)
    npe = CK.ElementTypeNPE[etype]
    if npe:
        nelems = len(earray.flat) // npe
    else:
        raise NotImplementedError('Oupss not implemented variable number of elems')
    start = r[1] + 1
    end = start + nelems - 1
    return numpy.array([start, end], dtype=numpy.int32)



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = 'zone structured'
diag = True
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone bad zonetype'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
zt = CGU.hasChildName(z, CGK.ZoneType_s)
zt[1] = CGU.setStringAsArray('Untruscutred')
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone bad zone size'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones(5, dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones(5, dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones(5, dtype='float64', order='F'))
zt = CGU.hasChildName(z, CGK.ZoneType_s)
zt[1] = CGU.setStringAsArray(CGK.Unstructured_s)
tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s, NPY.array([[1, 4]], 'i', order='F'), NPY.ones((4 * 4), dtype='int32'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone unstructured bad dims'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[4, 5, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones(4, dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones(4, dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones(4, dtype='float64', order='F'))
zt = CGU.hasChildName(z, CGK.ZoneType_s)
zt[1] = CGU.setStringAsArray(CGK.Unstructured_s)
tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s, NPY.array([[1, 5]], 'i', order='F'), NPY.ones((5 * 4), dtype='int32'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone structured bad dims'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 7, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone structured bad numpy array order'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 7, 9], [4, 6, 8], [0, 0, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 4, 0), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 4, 0), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 4, 0), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone unstructured no elements'
diag = False


def makeUnstTree(vertexsize, cellsize):
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, 'Base', 3, 3)
    s = NPY.array([[vertexsize, cellsize, 0]], dtype='int32', order='F')
    z = CGL.newZone(b, 'Zone', s, CGK.Unstructured_s)
    g = CGL.newGridCoordinates(z, 'GridCoordinates')
    d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    return (T, b, z)


vertexsize = 20
cellsize = 7
(T, b, z) = makeUnstTree(vertexsize, cellsize)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone unstructured correct ElementRange combination'
diag = True
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s,
                         NPY.array([[1, 3]], 'i', order='F'),
                         NPY.ones((3 * 4), dtype='int32'))
hexas = CGL.newElements(z, 'HEXAS', CGK.HEXA_8_s,
                        NPY.array([[4, cellsize]], 'i', order='F'),
                        NPY.ones((4 * 8), dtype='int32'))
element = CGL.newElements(z, 'NGON', CGK.NGON_n_s,
                          NPY.array([[cellsize + 1, cellsize + 6]], 'i', order='F'),
                          NPY.array([9, 9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9,
                                     9, 9, 9, 9], dtype='int32', order='F'),
                          NPY.array([0, 4, 9, 12, 17, 20, 24], dtype='int32', order='F')
                          )
hexas2 = CGL.newElements(z, 'HEXAS2', CGK.HEXA_8_s,
                         NPY.array([[cellsize + 7, cellsize + 8]], 'i', order='F'),
                         NPY.ones((2 * 8), dtype='int32'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone unstructured bad ElementRange combination #1'
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s,
                         NPY.array([[1, 3]], 'i', order='F'),
                         NPY.ones((3 * 4), dtype='int32'))
hexas = CGL.newElements(z, 'HEXAS', CGK.HEXA_8_s,
                        NPY.array([[3, 7]], 'i', order='F'),
                        NPY.ones((5 * 8), dtype='int32'))  # Bad combination of ElementRange
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone unstructured bad ElementRange combination #2'
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s,
                         NPY.array([[1, cellsize]], 'i', order='F'),
                         NPY.ones((cellsize * 4), dtype='int32'))
element = CGL.newElements(z, 'NGON', CGK.NGON_n_s,
                          NPY.array([[cellsize + 1 + 1, cellsize + 7 + 1]], 'i',
                                    order='F'),
                          NPY.array([9, 9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9, 9], dtype='int32', order='F'),
                          NPY.array([0, 4, 9, 12, 17, 20, 25, 29], dtype='int32', order='F')
                          )  # should be cellsize+1,cellsize+7
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone unstructured bad ElementRange combination #3'
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s, NPY.ones((cellsize * 4), dtype='int32'),
                         NPY.array([[1, cellsize]], 'i', order='F'))
element = CGL.newElements(z, 'NGON', CGK.NGON_n_s,
                          NPY.array([[cellsize, cellsize + 7 - 1]], 'i', order='F'),
                          NPY.array([9, 9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9, 9], dtype='int32', order='F'),
                          NPY.array([0, 4, 9, 12, 17, 20, 25, 29], dtype='int32', order='F')
                          )  # should be cellsize+1,cellsize+7
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone unstructured NFACE without NGON'
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s,
                         NPY.array([[1, cellsize]], 'i', order='F'),
                         NPY.ones((cellsize * 4), dtype='int32'))
element = CGL.newElements(z, 'NFACE', CGK.NFACE_n_s,
                          NPY.array([[cellsize + 1, cellsize + 5]], 'i', order='F'),
                          NPY.array([9, 9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9,
                                     9, 9, 9, 9, 9,
                                     9, 9, 9, 9], dtype='int32', order='F'),
                          NPY.array([0, 4, 9, 12, 17, 21], dtype='int32', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone structured full BC and GridConnect'
diag = True


def makeStTree():
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, '{Base}', 3, 3)
    z1 = CGL.newZone(b, '{Zone1}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
    g = CGL.newGridCoordinates(z1, 'GridCoordinates')
    d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
    z2 = CGU.copyNode(z1, '{Zone2}')
    b[2].append(z2)
    zgc = CGL.newZoneGridConnectivity(z1)
    gc = CGL.newGridConnectivity1to1(zgc, 'join1_2', '{Zone2}', NPY.array([[1, 1], [1, 4], [1, 9]]),
                                     NPY.array([[5, 5], [3, 7], [1, 9]]), NPY.array([-1, +2, +3]))
    zgc = CGL.newZoneGridConnectivity(z2)
    gc = CGL.newGridConnectivity1to1(zgc, 'join2_1', '{Zone1}', NPY.array([[5, 5], [3, 7], [1, 9]]),
                                     NPY.array([[1, 1], [1, 4], [1, 9]]), NPY.array([-1, +2, +3]))
    zbc = CGL.newZoneBC(z1)
    n = CGL.newBoundary(zbc, '{BC1_1}', [[5, 5], [1, 7], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC1_2}', [[1, 5], [1, 1], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC1_3}', [[1, 5], [7, 7], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC1_4}', [[1, 5], [1, 7], [1, 1]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC1_5}', [[1, 5], [1, 7], [9, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC1_6}', [[1, 1], [4, 7], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    zbc = CGL.newZoneBC(z2)
    n = CGL.newBoundary(zbc, '{BC2_1}', [[1, 1], [1, 7], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC2_2}', [[1, 5], [1, 1], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC2_3}', [[1, 5], [7, 7], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC2_4}', [[1, 5], [1, 7], [1, 1]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC2_5}', [[1, 5], [1, 7], [9, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(zbc, '{BC2_6}', [[5, 5], [1, 3], [1, 9]], btype=CGK.Null_s, family=None,
                        pttype=CGK.PointRange_s)
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    z = [z1, z2]
    return (T, b, z)


(T, b, z) = makeStTree()
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone structured uncomplete BC and GridConnect (warning)'
diag = False
(T, b, z) = makeStTree()
pth = CGU.getAllNodesByTypeOrNameList(z[0], ['Zone_t', 'ZoneBC_t', '{BC1_1}'])[0]
CGU.removeChildByName(CGU.getNodeByPath(z[0], CGU.getPathAncestor(pth)), '{BC1_1}')
pth = CGU.getAllNodesByTypeOrNameList(z[1], ['Zone_t', 'ZoneGridConnectivity_t', 'join2_1'])[0]
CGU.removeChildByName(CGU.getNodeByPath(z[1], CGU.getPathAncestor(pth)), 'join2_1')
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'zone structured doubly defined BC and GridConnect (warning)'
diag = False
(T, b, z) = makeStTree()
pth = CGU.getAllNodesByTypeOrNameList(z[0], ['Zone_t', 'ZoneBC_t'])[0]
zbc = CGU.getNodeByPath(z[0], pth)
n = CGL.newBoundary(zbc, '{BC1_1b}', [[5, 5], [1, 2], [1, 2]], btype=CGK.Null_s, family=None, pttype=CGK.PointRange_s)
g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
pth = CGU.getAllNodesByTypeOrNameList(z[1], ['Zone_t', 'ZoneGridConnectivity_t'])[0]
zgc = CGU.getNodeByPath(z[1], pth)
gc = CGL.newGridConnectivity1to1(zgc, 'join2_1b', '{Zone1}', NPY.array([[1, 1], [1, 2], [1, 2]]),
                                 NPY.array([[1, 1], [1, 4], [1, 9]]), NPY.array([-1, +2, +3]))
TESTS.append((tag, T, diag))



#  -------------------------------------------------------------------------
#  pyCGNS.NAV - Python package for CFD General Notation System - NAVigater
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnstypes as CGT
import CGNS.PAT.cgnskeywords as CGK
import CGNS.VAL.parse.messages as CGM
import CGNS.VAL.parse.generic

messagetable = (
    ('U0101', CGM.CHECK_WARN, 'No Zone in this Base'),
    ('U0102', CGM.CHECK_WARN, 'No Structured Zone found'),
    ('U0105', CGM.CHECK_FAIL, 'At least one structured Zone is required in the Base'),
    ('U0103', CGM.CHECK_WARN, 'No ReferenceState found at Base level'),
    ('U0104', CGM.CHECK_WARN, 'No ReferenceState found at Zone level'),
    ('U0107', CGM.CHECK_WARN, 'No FlowSolution# found for output definition'),
    ('U0108', CGM.CHECK_WARN, 'No FlowSolution#Init found for fields initialisation'),
    ('U0106', CGM.CHECK_FAIL, 'Transform is not right-handed (direct)'),
    ('U0109', CGM.CHECK_FAIL, 'Cannot handle such GridLocation [%s]'),
    ('U0110', CGM.CHECK_FAIL, 'Cannot handle such ElementType [%s]'),

    ('S0300', CGM.CHECK_FAIL, 'FamilyName is empty'),
    ('S0301', CGM.CHECK_FAIL, 'Reference to unknown family [%s]'),
    ('S0302', CGM.CHECK_FAIL, 'Reference to unknown additional family [%s]'),

    ('S0602', CGM.CHECK_FAIL, 'Zone has no GridCoordinates'),
    ('S0603', CGM.CHECK_FAIL, 'No GridCoordinates_t of name GridCoordinates in zone'),
    ('S0604', CGM.CHECK_WARN, 'ZoneBC has no BC'),
)

USER_MESSAGES = {}
for (k, l, m) in messagetable:
    USER_MESSAGES[k] = (l, m)


# -----------------------------------------------------------------------------
class CGNS_VAL_USER_Checks(CGNS.VAL.parse.generic.GenericParser):
    def __init__(self, log):
        CGNS.VAL.parse.generic.GenericParser.__init__(self, log)
        self.log.addMessages(USER_MESSAGES)

    # --------------------------------------------------------------------
    def Zone_t(self, pth, node, parent, tree, log):
        rs = CGM.CHECK_OK
        zt = CGU.hasChildName(node, CGK.ZoneType_s)
        zv = []
        if zt is not None:
            if CGU.stringValueMatches(zt, CGK.Structured_s):
                cd = self.context[CGK.CellDimension_s][pth]
                self.context[CGK.IndexDimension_s][pth] = cd
            elif CGU.stringValueMatches(zt, CGK.Unstructured_s):
                self.context[CGK.IndexDimension_s][pth] = 1
            shp = (self.context[CGK.IndexDimension_s][pth], 3)
            if CGU.getShape(node) != shp:
                rs = log.push(pth, 'S0009', CGU.getShape(node))
            elif CGU.stringValueMatches(zt, CGK.Structured_s):
                zd = node[1]
                for nd in range(self.context[CGK.IndexDimension_s][pth]):
                    zv.append(zd[nd][0])
                    if ((zd[nd][1] != zd[nd][0] - 1) or (zd[nd][2] != 0)):
                        rs = log.push(pth, 'S0010')
            else:
                zv.append(node[1][0][0])
            self.context[CGK.VertexSize_s][pth] = tuple(zv)
        if CGU.hasChildNodeOfType(node, CGK.FamilyName_ts):
            basepath = [CGK.CGNSTree_ts, parent[0], node[0]]
            searchpath = basepath + [CGK.FamilyName_ts]
            famlist1 = CGU.getAllNodesByTypeOrNameList(tree, searchpath)
            searchpath = basepath + [CGK.AdditionalFamilyName_ts]
            famlist2 = CGU.getAllNodesByTypeOrNameList(tree, searchpath)
            for (famlist, diagmessage) in ((famlist1, 'S0301'),
                                           (famlist2, 'S0302')):
                for fampath in famlist:
                    famdefinition = CGU.getNodeByPath(tree, fampath)
                    if (famdefinition[1] is None):
                        rs = log.push(pth, 'S0300')
                    else:
                        famtarget = famdefinition[1].tostring().rstrip()
                        famtargetpath = "/%s/%s" % (parent[0], famtarget)
                        if (famtargetpath not in self.context):
                            famtargetnode = CGU.getNodeByPath(tree, famtargetpath)
                            if (famtargetnode is None):
                                rs = log.push(pth, diagmessage, famtarget)
                            else:
                                self.context[famtargetpath][pth] = True
        if (not CGU.hasChildType(node, CGK.GridCoordinates_ts)):
            rs = log.push(pth, 'S0602')
        elif (not CGU.hasChildName(node, CGK.GridCoordinates_s)):
            rs = log.push(pth, 'S0603')
        if (not CGU.hasChildType(node, CGK.ZoneBC_ts)):
            rs = log.push(pth, 'S0604')
        return rs

    # --------------------------------------------------------------------
    def ZoneType_t(self, pth, node, parent, tree, log):
        rs = CGM.CHECK_OK
        if (not CGU.stringValueInList(node, CGK.ZoneType_l)):
            rs = log.push(pth, 'S0101')
        return rs

    # --------------------------------------------------------------------
    def CGNSBase_t(self, pth, node, parent, tree, log):
        rs = CGM.CHECK_OK
        (cd, pd) = (0, 0)
        if not CGU.hasChildNodeOfType(node, CGK.Zone_ts):
            rs = log.push(pth, 'U0101')
        else:
            target = [CGK.CGNSTree_ts, node[0], CGK.Zone_ts, CGK.ZoneType_s]
            plist = CGU.getAllNodesByTypeOrNameList(tree, target)
            found = False
            for p in plist:
                if (CGU.stringValueMatches(CGU.getNodeByPath(tree, p),
                                           CGK.Structured_s)):
                    found = True
            if not found:
                rs = log.push(pth, 'U0102')
        if not CGU.hasChildNodeOfType(node, CGK.ReferenceState_ts):
            rs = log.push(pth, 'U0103')
        if CGU.getShape(node) != (2,):
            rs = log.push(pth, 'S0009', CGU.getShape(node))
        else:
            cd = node[1][0]
            pd = node[1][1]
            allowedvalues = ((1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3))
            if (cd, pd) not in allowedvalues:
                rs = log.push(pth, 'S0010', (cd, pd))
        self.context[CGK.CellDimension_s] = cd
        self.context[CGK.PhysicalDimension_s] = pd
        return rs

    # --------------------------------------------------------------------
    def GridLocation_t(self, pth, node, parent, tree, log):
        rs = CGM.CHECK_OK
        val = node[1].tostring()
        if val not in [CGK.Vertex_s, CGK.CellCenter_s, CGK.FaceCenter_s]:
            rs = log.push(pth, 'U0109', val)
        return rs

# -----



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  ---------------------------------------------------------------------------
#
from __future__ import unicode_literals
import CGNS.PAT.cgnslib as C
import CGNS.PAT.cgnserrors as E
import CGNS.PAT.cgnskeywords as K
import numpy as N

#
data = C.newBCProperty(None)
C.newDescriptor(data, '{Descriptor}')
C.newUserDefinedData(data, '{UserDefinedData}')
#for n in data[2]:
#    if n[0] == K.Area_s:
#        C.newArea(n)
#    if (n[0] == K.WallFunction_s):
#        C.newDescriptor(n, '{Descriptor}')
#        C.newUserDefinedData(n, '{UserDefinedData}')
# #
status = '9.6'
comment = 'Full SIDS with all optionals'
pattern = [data, status, comment]
#



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System - VALidater
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = 'indexarray'
diag = True


def makeCorrectTree(vertexsize, cellsize):
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, 'Base', 3, 3)
    s = NPY.array([[vertexsize, cellsize, 0]], dtype='int32', order='F')
    z = CGL.newZone(b, 'Zone', s, CGK.Unstructured_s)
    g = CGL.newGridCoordinates(z, 'GridCoordinates')
    d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    tetras = CGL.newElements(z, 'TETRAS', CGK.TETRA_4_s, NPY.array([[1, cellsize]], 'i', order='F'),
                             NPY.ones((cellsize * 4), dtype='int32'))
    zbc = CGL.newZoneBC(z)
    return (T, b, z, zbc)


vertexsize = 20
cellsize = 7
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray bad parent'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
i = n[2][0]
z[2].append(CGU.copyNode(i, 'PointList'))  # unauthorized parent node
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray bad name'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
i = n[2][0]
i[0] = 'ElementList'  # unauthorized name
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray bad data type'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
i = n[2][0]
i[1] = NPY.ones((1, cellsize), dtype='float64', order='F')  # unauthorized data type
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray bad child'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
i = n[2][0]
CGL.newDataArray(i, '{DataArray}')  # unauthorized child
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray element index out of range'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(2, cellsize + 2)], btype=CGK.Null_s, family=None,
                    pttype=CGK.PointList_s)  # element index out of range
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray element appear several times (Warning)'
diag = True
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [[1, 1, 1, 1, 1, 1, 1]], btype=CGK.Null_s, family=None,
                    pttype=CGK.PointList_s)  # values appear several times in list
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray gridlocation vertex'
diag = True
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [[1, 2, 3, 4, vertexsize]], btype=CGK.Null_s, family=None,
                    pttype=CGK.PointList_s)  # element index out of range
g = CGL.newGridLocation(n, value=CGK.Vertex_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray gridlocation vertex index out of range'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [[1, 2, 0, 4, vertexsize + 1]], btype=CGK.Null_s, family=None,
                    pttype=CGK.PointList_s)  # element index out of range
g = CGL.newGridLocation(n, value=CGK.Vertex_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray gridlocation face'
diag = True
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
ntris = 12
tris = CGL.newElements(z, 'TRIS', CGK.TRI_3_s, NPY.array([[cellsize + 1, cellsize + ntris]], 'i', order='F'),
                       NPY.ones((ntris * 3), dtype='int32'))
n = CGL.newBoundary(zbc, 'BC', [range(cellsize + 1, cellsize + ntris + 1)], btype=CGK.Null_s, family=None,
                    pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray gridlocation face index out of range #1'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
ntris = 12
tris = CGL.newElements(z, 'TRIS', CGK.TRI_3_s, NPY.array([[cellsize + 1, cellsize + ntris]], 'i', order='F'),
                       NPY.ones((ntris * 3), dtype='int32'))
n = CGL.newBoundary(zbc, 'BC', [range(cellsize + 1, cellsize + ntris + 2)], btype=CGK.Null_s, family=None,
                    pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray gridlocation face index out of range #2'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
ntris = 12
tris = CGL.newElements(z, 'TRIS', CGK.TRI_3_s, NPY.array([[cellsize + 1, cellsize + ntris]], 'i', order='F'),
                       NPY.ones((ntris * 3), dtype='int32'))
n = CGL.newBoundary(zbc, 'BC', [range(cellsize, cellsize + ntris + 1)], btype=CGK.Null_s, family=None,
                    pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray InwardNormalList'
diag = True
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
inl = CGL.newIndexArray(n, CGK.InwardNormalList_s, value=NPY.ones([3, cellsize]))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray InwardNormalList bad shape #1'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
inl = CGL.newIndexArray(n, CGK.InwardNormalList_s, value=NPY.ones([2, cellsize]))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray InwardNormalList bad shape #2'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
inl = CGL.newIndexArray(n, CGK.InwardNormalList_s, value=NPY.ones([3, cellsize + 1]))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray InwardNormalList bad shape #3'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
inl = CGL.newIndexArray(n, CGK.InwardNormalList_s, value=NPY.ones([cellsize]))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'indexarray InwardNormalList bad shape #4'
diag = False
(T, b, z, zbc) = makeCorrectTree(vertexsize, cellsize)
n = CGL.newBoundary(zbc, 'BC', [range(1, cellsize + 1)], btype=CGK.Null_s, family=None, pttype=CGK.PointList_s)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
CGU.nodeDelete(T, CGU.getNodeByPath(n, 'BC/' + CGK.PointList_s))
inl = CGL.newIndexArray(n, CGK.InwardNormalList_s, value=NPY.ones([2, cellsize]))
TESTS.append((tag, T, diag))
